/****************************************************************************
 *
 * Copyright (c) 2001-2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

#ifndef _CALCMDD_H
#define _CALCMDD_H

#include <connio.h>
#include <mdb.h>
#include <management.h>
#include <msgapi.h>
#include <nmap.h>
#include <nmlib.h>
#include <calcmd.h>

#define PRODUCT_SHORT_NAME "hulacalcmd.nlm"

#define CONNECTION_TIMEOUT (15 * 60)

typedef enum _CalCmdStates {
    CALCMD_STATE_STARTING = 0, 
    CALCMD_STATE_INITIALIZING, 
    CALCMD_STATE_LOADING, 
    CALCMD_STATE_RUNNING, 
    CALCMD_STATE_RELOADING, 
    CALCMD_STATE_UNLOADING, 
    CALCMD_STATE_STOPPING, 
    CALCMD_STATE_DONE, 

    CALCMD_STATE_MAX_STATES
} CalCmdStates;

typedef enum _CalCmdFlags {
    CALCMD_FLAG_UNUSED = (1 << 0), 
} CalCmdFlags;

typedef enum _CalCmdClientFlags {
    CALCMD_CLIENT_FLAG_NEW = (1 << 0), 
    CALCMD_CLIENT_FLAG_WAITING = (1 << 1), 
    CALCMD_CLIENT_FLAG_EXITING = (1 << 2)
} CalCmdClientFlags;

typedef struct {
    CalCmdClientFlags flags;

    Connection *conn;

    char *envelope;
    char line[CONN_BUFSIZE + 1];
    char command[CONN_BUFSIZE + 1];
    char dn[MDB_MAX_OBJECT_CHARS + 1];

    char sender[CONN_BUFSIZE + 1];
    char recipient[CONN_BUFSIZE + 1];
    char user[CONN_BUFSIZE + 1];
    char calreq[CONN_BUFSIZE + 1];

} CalCmdClient;

typedef struct _CalCmdGlobals {
    CalCmdStates state;
    CalCmdFlags flags;

    char officialName[MAXEMAILNAMESIZE + 1];

    struct {
        XplSemaphore semaphore;

        struct {
            XplSemaphore todo;

            XplAtomic maximum;
            XplAtomic active;
            XplAtomic idle;

            Connection *head;
            Connection *tail;
        } worker;

        struct {
            BOOL enable;

            ConnSSLConfiguration config;

            SSL_CTX *context;

            Connection *conn;
        } ssl;

        Connection *conn;

        void *pool;

        time_t sleepTime;

        unsigned long queue;

        char address[80];
        char hash[NMAP_HASH_SIZE];
    } nmap;

    struct {
        MDBHandle directory;

        void *logging;
    } handle;

    struct {
        XplThreadID main;
        XplThreadID group;
    } id;

    struct {
        XplSemaphore main;
        XplSemaphore shutdown;
    } sem;

    struct {
        XplAtomic active;
    } server;

    struct {
        XplAtomic called;
    } stats;
} CalCmdGlobals;

extern CalCmdGlobals CalCmd;

/* management.c */
ManagementVariables *GetCalCmdManagementVariables(void);
int GetCalCmdManagementVariablesCount(void);
ManagementCommands *GetCalCmdManagementCommands(void);
int GetCalCmdManagementCommandsCount(void);

#endif /* _CALCMDD_H */
