
/****************************************************************************
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 ****************************************************************************/

#include <config.h>

#define N_PLAT_NLM
#define N_IAPX386

/* Product defines */
#define PRODUCT_SHORT_NAME "imapd.nlm"
#define PRODUCT_NAME   "Hula IMAP4 Agent"
#define PRODUCT_DESCRIPTION "Provides IMAP4 access to the Hula mail-store. (IMAP4 = Internet Mail Access Protocol, Rev4, RFC2060)"
#define PRODUCT_VERSION  "$Revision: 1.8 $"

#define MAX_THREAD_LOAD  50
#define MAX_FAILED_LOGINS  3
#undef DEBUG_FETCH
#define NMLOGID_H_NEED_LOGGING_KEY
#define NMLOGID_H_NEED_LOGGING_CERT
#include <xpl.h>

#include <logger.h>

#include <mdb.h>
#include <msgapi.h>
#include <cmlib.h>

/* Management Client Header Include */
#include <management.h>
#include "imapd.h"

ImapGlobal Imap;

static ProtocolCommand ImapProtocolCommands[] = {
    { IMAP_COMMAND_CAPABILITY, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_CAPABILITY) - 1, ImapCommandCapability, NULL, NULL },
    { IMAP_COMMAND_NOOP, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_NOOP) - 1, ImapCommandNoop, NULL, NULL },
    { IMAP_COMMAND_LOGOUT, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_LOGOUT) - 1, ImapCommandLogout, NULL, NULL },
    { IMAP_COMMAND_STARTTLS, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_STARTTLS) - 1, ImapCommandStartTls, NULL, NULL },
    { IMAP_COMMAND_AUTHENTICATE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_AUTHENTICATE) - 1, ImapCommandAuthenticate, NULL, NULL },
    { IMAP_COMMAND_LOGIN, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_LOGIN) - 1, ImapCommandLogin, NULL, NULL },
    { IMAP_COMMAND_SELECT, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_SELECT) - 1, ImapCommandSelect, NULL, NULL },
    { IMAP_COMMAND_EXAMINE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_EXAMINE) - 1, ImapCommandExamine, NULL, NULL },
    { IMAP_COMMAND_CREATE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_CREATE) - 1, ImapCommandCreate, NULL, NULL },
    { IMAP_COMMAND_DELETE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_DELETE) - 1, ImapCommandDelete, NULL, NULL },
    { IMAP_COMMAND_RENAME, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_RENAME) - 1, ImapCommandRename, NULL, NULL },
    { IMAP_COMMAND_SUBSCRIBE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_SUBSCRIBE) - 1, ImapCommandSubscribe, NULL, NULL },
    { IMAP_COMMAND_UNSUBSCRIBE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_UNSUBSCRIBE) - 1, ImapCommandUnsubscribe, NULL, NULL },
    { IMAP_COMMAND_LIST, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_LIST) - 1, ImapCommandList, NULL, NULL },
    { IMAP_COMMAND_LSUB, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_LSUB) - 1, ImapCommandLsub, NULL, NULL },
    { IMAP_COMMAND_STATUS, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_STATUS) - 1, ImapCommandStatus, NULL, NULL },
    { IMAP_COMMAND_APPEND, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_APPEND) - 1, ImapCommandAppend, NULL, NULL },
    { IMAP_COMMAND_CHECK, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_CHECK) - 1, ImapCommandCheck, NULL, NULL },
    { IMAP_COMMAND_CLOSE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_CLOSE) - 1, ImapCommandClose, NULL, NULL },
    { IMAP_COMMAND_EXPUNGE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_EXPUNGE) - 1, ImapCommandExpunge, NULL, NULL },
    { IMAP_COMMAND_SEARCH, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_SEARCH) - 1, ImapCommandSearch, NULL, NULL },
    { IMAP_COMMAND_UID_SEARCH, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_UID_SEARCH) - 1, ImapCommandSearch, NULL, NULL },
    { IMAP_COMMAND_FETCH, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_FETCH) - 1, ImapCommandFetch, NULL, NULL },
    { IMAP_COMMAND_UID_FETCH, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_UID_FETCH) - 1, ImapCommandUidFetch, NULL, NULL },
    { IMAP_COMMAND_STORE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_STORE) - 1, ImapCommandStore, NULL, NULL },
    { IMAP_COMMAND_UID_STORE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_UID_STORE) - 1, ImapCommandUidStore, NULL, NULL },
    { IMAP_COMMAND_COPY, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_COPY) - 1, ImapCommandCopy, NULL, NULL },
    { IMAP_COMMAND_UID_COPY, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_UID_COPY) - 1, ImapCommandUidCopy, NULL, NULL },
    { IMAP_COMMAND_SETACL, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_SETACL) - 1, ImapCommandSetAcl, NULL, NULL },
    { IMAP_COMMAND_DELETEACL, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_DELETEACL) - 1, ImapCommandDeleteAcl, NULL, NULL },
    { IMAP_COMMAND_GETACL, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_GETACL) - 1, ImapCommandGetAcl, NULL, NULL },
    { IMAP_COMMAND_LISTRIGHTS, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_LISTRIGHTS) - 1, ImapCommandListRights, NULL, NULL },
    { IMAP_COMMAND_MYRIGHTS, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_MYRIGHTS) - 1, ImapCommandMyRights, NULL, NULL },
    { IMAP_COMMAND_SETQUOTA, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_SETQUOTA) - 1, ImapCommandSetQuota, NULL, NULL },
    { IMAP_COMMAND_GETQUOTA, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_GETQUOTA) - 1, ImapCommandGetQuota, NULL, NULL },
    { IMAP_COMMAND_GETQUOTAROOT, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_GETQUOTAROOT) - 1, ImapCommandGetQuotaRoot, NULL, NULL },
    { IMAP_COMMAND_NAMESPACE, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_NAMESPACE) - 1, ImapCommandNameSpace, NULL, NULL },
    { IMAP_COMMAND_PROXYAUTH, IMAP_HELP_NOT_DEFINED, sizeof(IMAP_COMMAND_PROXYAUTH) - 1, ImapCommandProxyAuth, NULL, NULL },
    
    { NULL, NULL, 0, NULL, NULL, NULL }
};

#undef DEBUG

#if !defined(DEBUG)
#define GetIMAPConnection()   MemPrivatePoolGetEntry(IMAPConnectionPool)
#else
#define GetIMAPConnection()   MemPrivatePoolGetEntryDebug(IMAPConnectionPool, __FILE__, __LINE__)
#endif

void    *IMAPConnectionPool  = NULL;

/* Management Client Declarations */
BOOL ReadIMAPVariable(unsigned int Variable, unsigned char *Data, size_t *DataLength);
BOOL WriteIMAPVariable(unsigned int Variable, unsigned char *Data, size_t DataLength);

ManagementVariables IMAPManagementVariables[] = {
    /* 0 #define    PRODUCT_VERSION       */ { DMCMV_REVISIONS,     DMCMV_REVISIONS_HELP,    ReadIMAPVariable, NULL }, 
    /* 1 unsigned long  Imap.client.threads.max       */ { DMCMV_MAX_CONNECTION_COUNT, DMCMV_MAX_CONNECTION_COUNT_HELP, ReadIMAPVariable, WriteIMAPVariable }, 
    /* 2 XplAtomic   IMAPServerThreads       */ { DMCMV_SERVER_THREAD_COUNT,  DMCMV_SERVER_THREAD_COUNT_HELP, ReadIMAPVariable, NULL }, 
    /* 3 XplAtomic   Imap.client.threads.inUse       */ { DMCMV_CONNECTION_COUNT,   DMCMV_CONNECTION_COUNT_HELP,  ReadIMAPVariable, NULL }, 
    /* 4 XplAtomic   Imap.client.threads.idle       */ { DMCMV_IDLE_CONNECTION_COUNT,   DMCMV_IDLE_CONNECTION_COUNT_HELP,  ReadIMAPVariable, NULL }, 
    /* 5 BOOL     Imap.command.capability.acl.enabled        */ { IMAPMV_ACL_ENABLED,    IMAPMV_ACL_ENABLED_HELP,   ReadIMAPVariable, NULL }, 
    /* 6 BOOL     Imap.server.disabled      */ { DMCMV_RECEIVER_DISABLED,  DMCMV_RECEIVER_DISABLED_HELP,  ReadIMAPVariable, WriteIMAPVariable }, 
    /* 7 unsigned char  Imap.server.hostname[256]        */ { DMCMV_HOSTNAME,     DMCMV_HOSTNAME_HELP,     ReadIMAPVariable, NULL }, 
    /* 8 unsigned char  Imap.server.postmaster[MAXEMAILNAMESIZE + 1]  */ { DMCMV_POSTMASTER,     DMCMV_POSTMASTER_HELP,    ReadIMAPVariable, NULL }, 
    /* 9 XplAtomic   Imap.client.served    */ { DMCMV_TOTAL_CONNECTIONS,  DMCMV_TOTAL_CONNECTIONS_HELP,  ReadIMAPVariable, WriteIMAPVariable }, 
    /* 10 XplAtomic   Imap.client.badPassword     */ { DMCMV_BAD_PASSWORD_COUNT,  DMCMV_BAD_PASSWORD_COUNT_HELP, ReadIMAPVariable, WriteIMAPVariable }, 
    /* 11 unsigned char DMC Version              */ { DMCMV_VERSION,      DMCMV_VERSION_HELP,     ReadIMAPVariable, NULL }, 
};

static void
FreeReturnValueIndex(unsigned long *index)
{
    MemFree(index);
}

static long *
CreateReturnValueIndex()
{
    long status;
    unsigned long count;
    unsigned long i;
    long *index;


    index = MemMalloc(STATUS_MAX * sizeof(unsigned long *));
    if (index) {


        /* find the number of strings */
        count = 0;

        do {
            if (ImapErrorStrings[count].formatString != NULL) {
                count++;
                continue;
            }
            break;
        } while (TRUE);

        for (status = 0; status < STATUS_MAX; status++) {
            index[status] = -1;
            for (i = 0; i < count; i++) {
                if (ImapErrorStrings[i].id != status) {
                    continue;
                }
                index[status] = i;
                break;
            }
        }
        return(index);
    }
    return(NULL);
}

static BOOL 
IMAPShutdown(unsigned char *Arguments, unsigned char **Response, BOOL *CloseConnection)
{
    XplThreadID id;

    if (Response) {
        if (!Arguments) {
            if (Imap.server.conn->socket != -1) {
                *Response = MemStrdup("Shutting down.\r\n");
                if (*Response) {
                    id = XplSetThreadGroupID(Imap.server.threadGroupId);

                    Imap.exiting = TRUE;

                    if (Imap.server.conn) {
                        ConnClose(Imap.server.conn, 1);
                        Imap.server.conn = NULL;
                    }

                    XplSignalLocalSemaphore(Imap.server.shutdownSemaphore);

                    if (CloseConnection) {
                        *CloseConnection = TRUE;
                    }

                    XplSetThreadGroupID(id);
                }
            } else if (Imap.exiting) {
                *Response = MemStrdup("Shutdown in progress.\r\n");
            } else {
                *Response = MemStrdup("Unknown shutdown state.\r\n");
            }

            if (*Response) {
                return(TRUE);
            }

            return(FALSE);
        }

        *Response = MemStrdup("Arguments not allowed.\r\n");
        return(TRUE);
    }

    return(FALSE);
}

static BOOL 
IMAPDMCCommandHelp(unsigned char *Arguments, unsigned char **Response, BOOL *CloseConnection)
{
    BOOL responded = FALSE;

    if (Response) {
        if (Arguments) {
            switch(toupper(Arguments[0])) {
            case 'M': {
                if (XplStrNCaseCmp(Arguments, DMCMC_DUMP_MEMORY_USAGE, sizeof(DMCMC_DUMP_MEMORY_USAGE) - 1) == 0) {
                    if ((*Response = MemStrdup(DMCMC_DUMP_MEMORY_USAGE_HELP)) != NULL) {
                        responded = TRUE;
                    }

                    break;
                }
            }

            case 'C': {
                if (XplStrNCaseCmp(Arguments, DMCMC_CONN_TRACE_USAGE, sizeof(DMCMC_CONN_TRACE_USAGE) - 1) == 0) {
                    if ((*Response = MemStrdup(DMCMC_CONN_TRACE_USAGE_HELP)) != NULL) {
                        responded = TRUE;
                    }

                    break;
                }
            }

            case 'S': {
                if (XplStrNCaseCmp(Arguments, DMCMC_STATS, sizeof(DMCMC_STATS) - 1) == 0) {
                    if ((*Response = MemStrdup(DMCMC_STATS_HELP)) != NULL) {
                        responded = TRUE;
                    }

                    break;
                } else if (XplStrNCaseCmp(Arguments, DMCMC_SHUTDOWN, sizeof(DMCMC_SHUTDOWN) - 1) == 0) {
                    if ((*Response = MemStrdup(DMCMC_SHUTDOWN_HELP)) != NULL) {
                        responded = TRUE;
                    }

                    break;
                }
            }

            default: {
                break;
            }
            }
        } else if ((*Response = MemStrdup(DMCMC_HELP_HELP)) != NULL) {
            responded = TRUE;
        }

        if (responded || ((*Response = MemStrdup(DMCMC_UNKOWN_COMMAND)) != NULL)) {
            return(TRUE);
        }
    }

    return(FALSE);
}

static BOOL 
SendIMAPStatistics(unsigned char *Arguments, unsigned char **Response, BOOL *CloseConnection)
{
    MemStatistics poolStats;

    if (!Arguments && Response) {
        memset(&poolStats, 0, sizeof(MemStatistics));

        *Response = MemMalloc(sizeof(PRODUCT_NAME)                             /* Long Name */
                              + sizeof(PRODUCT_SHORT_NAME) /* Short Name */
                              + 10    /* PRODUCT_MAJOR_VERSION */
                              + 10    /* PRODUCT_MINOR_VERSION */
                              + 10    /* PRODUCT_LETTER_VERSION */
                              + 10    /* Connection Pool Alloc Count */
                              + 10    /* Connection Pool Memory Usage */
                              + 10    /* Connection Pool Pitches */
                              + 10    /* Connection Pool Strikes */
                              + 10    /* DMCMV_SERVER_THREAD_COUNT */
                              + 10     /* DMCMV_CONNECTION_COUNT */
                              + 10     /* DMCMV_IDLE_CONNECTION_COUNT */
                              + 10    /* DMCMV_MAX_CONNECTION_COUNT */
                              + 10    /* DMCMV_TOTAL_CONNECTIONS */
                              + 10    /* DMCMV_BAD_PASSWORD_COUNT */
                              + 28);    /* Formatting */

        MemPrivatePoolStatistics(IMAPConnectionPool, &poolStats);

        if (*Response) {
            sprintf(*Response, "%s (%s: v%d.%d.%d)\r\n%lu:%lu:%lu:%lu:%d:%d:%d:%lu:%d:%d\r\n", 
                    PRODUCT_NAME, 
                    PRODUCT_SHORT_NAME, 
                    PRODUCT_MAJOR_VERSION, 
                    PRODUCT_MINOR_VERSION, 
                    PRODUCT_LETTER_VERSION, 
                    poolStats.totalAlloc.count, 
                    poolStats.totalAlloc.size, 
                    poolStats.pitches, 
                    poolStats.strikes, 
                    XplSafeRead(Imap.server.active), 
                    XplSafeRead(Imap.client.threads.inUse), 
                    XplSafeRead(Imap.client.threads.idle), 
                    Imap.client.threads.max, 
                    XplSafeRead(Imap.client.served), 
                    XplSafeRead(Imap.client.badPassword));

            return(TRUE);
        }

        if ((*Response = MemStrdup("Out of memory.\r\n")) != NULL) {
            return(TRUE);
        }
    } else if ((Arguments) && ((*Response = MemStrdup("Arguments not allowed.\r\n")) != NULL)) {
        return(TRUE);
    }

    return(FALSE);
}

ManagementCommands IMAPManagementCommands[] = {
    /* 0 HELP[ <IMAPCommand>]     */ { DMCMC_HELP,        IMAPDMCCommandHelp }, 
    /* 1 SHUTDOWN         */ { DMCMC_SHUTDOWN,      IMAPShutdown },
    /* 2 STATS          */ { DMCMC_STATS,       SendIMAPStatistics }, 
    /* 3 MEMORY         */ { DMCMC_DUMP_MEMORY_USAGE,   ManagementMemoryStats  }, 
    /* 4 CONNTRACEFLAGS */ { DMCMC_CONN_TRACE_USAGE,   ManagementConnTrace  }, 
};

/* Management Client Read Function */
BOOL ReadIMAPVariable(unsigned int Variable, unsigned char *Data, size_t *DataLength)
{
    unsigned long count;
    unsigned char *ptr;

    switch (Variable) {
    case 0: { /* #define    PRODUCT_VERSION       */
        unsigned char version[30];

        PVCSRevisionToVersion(PRODUCT_VERSION, version);
        count = strlen(version) + 11;

        if (Data && (*DataLength > count)) {
            ptr = Data;

            PVCSRevisionToVersion(PRODUCT_VERSION, version);
            ptr += sprintf(ptr, "imapd.c: %s\r\n", version);

            *DataLength = ptr - Data;
        } else {
            *DataLength = count;
        }

        break;
    }

    case 1: { /* unsigned long  Imap.client.threads.max       */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", Imap.client.threads.max);
        }

        *DataLength = 12;
        break;
    }

    case 2: { /* XplAtomic   IMAPServerThreads       */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", (long unsigned int)XplSafeRead(Imap.server.active));
        }

        *DataLength = 12;
        break;
    }

    case 3: { /* XplAtomic   Imap.client.threads.inUse       */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", (long unsigned int)XplSafeRead(Imap.client.threads.inUse));
        }

        *DataLength = 12;
        break;
    }

    case 4: { /* XplAtomic   Imap.client.threads.idle       */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", (long unsigned int)XplSafeRead(Imap.client.threads.idle));
        }

        *DataLength = 12;
        break;
    }

    case 5: { /* BOOL     Imap.command.capability.acl.enabled        */
        if (Imap.command.capability.acl.enabled == FALSE) {
            ptr = "FALSE\r\n";
            count = 7;
        } else {
            ptr = "TRUE\r\n";
            count = 6;
        }

        if (Data && (*DataLength > count)) {
            strcpy(Data, ptr);
        }

        *DataLength = count;
        break;
    }

    case 6: { /* BOOL     Imap.server.disabled      */
        if (Imap.server.disabled == FALSE) {
            ptr = "FALSE\r\n";
            count = 7;
        } else {
            ptr = "TRUE\r\n";
            count = 6;
        }

        if (Data && (*DataLength > count)) {
            strcpy(Data, ptr);
        }

        *DataLength = count;
        break;
    }

    case 7: { /* unsigned char  Imap.server.hostname[256]        */
        count = strlen(Imap.server.hostname) + 2;
        if (Data && (*DataLength > count)) {
            sprintf(Data, "%s\r\n", Imap.server.hostname);
        }

        *DataLength = count;
        break;
    }

    case 8: { /* unsigned char  Imap.server.postmaster[MAXEMAILNAMESIZE + 1]  */
        count = strlen(Imap.server.postmaster) + 2;
        if (Data && (*DataLength > count)) {
            sprintf(Data, "%s\r\n", Imap.server.postmaster);
        }

        *DataLength = count;
        break;
    }

    case 9: { /* XplAtomic   Imap.client.served    */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", (long unsigned int)XplSafeRead(Imap.client.served));
        }

        *DataLength = 12;
        break;
    }

    case 10: { /* XplAtomic   Imap.client.badPassword     */
        if (Data && (*DataLength > 12)) {
            sprintf(Data, "%010lu\r\n", (long unsigned int)XplSafeRead(Imap.client.badPassword));
        }

        *DataLength = 12;
        break;
    }

    case 11: { /* unsigned char */
        DMC_REPORT_PRODUCT_VERSION(Data, *DataLength);
        break;
    }
    }

    return(TRUE);
}

/* Management Client Write Function */
BOOL WriteIMAPVariable(unsigned int Variable, unsigned char *Data, size_t DataLength)
{
    unsigned char *ptr;
    unsigned char *ptr2;
    BOOL    result = TRUE;

    if (!Data || !DataLength) {
        return(FALSE);
    }

    switch (Variable) {
    case 1: { /* unsigned long  Imap.client.threads.max       */
        ptr = strchr(Data, '\n');
        if (ptr) {
            *ptr = '\0';
        }

        ptr2 = strchr(Data, '\r');
        if (ptr2) {
            ptr2 = '\0';
        }

        Imap.client.threads.max = atol(Data);

        if (ptr) {
            *ptr = '\n';
        }

        if (ptr2) {
            *ptr2 = 'r';
        }

        break;
    }

    case 6: { /* BOOL     Imap.server.disabled      */
        if ((toupper(Data[0]) == 'T') || (atol(Data) != 0)) {
            Imap.server.disabled = TRUE;
        } else if ((toupper(Data[0] == 'F')) || (atol(Data) == 0)) {
            Imap.server.disabled = FALSE;
        } else {
            result = FALSE;
        }

        break;
    }

    case 9: { /* XplAtomic   Imap.client.served    */
        ptr = strchr(Data, '\n');
        if (ptr) {
            *ptr = '\0';
        }

        ptr2 = strchr(Data, '\r');
        if (ptr2) {
            ptr2 = '\0';
        }

        XplSafeWrite(Imap.client.served, atol(Data));

        if (ptr) {
            *ptr = '\n';
        }

        if (ptr2) {
            *ptr2 = 'r';
        }

        break;
    }

    case 10: { /* XplAtomic   Imap.client.badPassword     */
        ptr = strchr(Data, '\n');
        if (ptr) {
            *ptr = '\0';
        }

        ptr2 = strchr(Data, '\r');
        if (ptr2) {
            ptr2 = '\0';
        }

        XplSafeWrite(Imap.client.badPassword, atol(Data));

        if (ptr) {
            *ptr = '\n';
        }

        if (ptr2) {
            *ptr2 = 'r';
        }

        break;
    }

    case 0: /* #define    PRODUCT_VERSION       */
    case 2: /* XplAtomic   IMAPServerThreads       */
    case 3: /* XplAtomic   Imap.client.threads.inUse       */
    case 4: /* XplAtomic   Imap.client.threads.idle       */
    case 5: /* BOOL     Imap.command.capability.acl.enabled        */
    case 7: /* unsigned char  Imap.server.hostname[256]        */
    case 8: /* unsigned char  Imap.server.postmaster[MAXEMAILNAMESIZE + 1]  */
    case 11: /* DMC Version */
    default: {
        result = FALSE;
        break;
    }
    }

    return(result);
}

static BOOL 
IMAPConnectionAllocCB(void *Buffer, void *ClientData)
{
    register ImapClient *c = (ImapClient *)Buffer;

    memset(c, 0, sizeof(ImapClient));

#if !defined(DEBUG)
    c->Command = (unsigned char *)MemMalloc(BUFSIZE + 1);
#else
    c->Command = (unsigned char *)MemDebugMalloc(BUFSIZE + 1, __FILE__, __LINE__);
#endif

    if (c->Command != NULL) {
        c->Command[0] = '\0';

        c->state = STATE_FRESH;
        c->CommandLen = BUFSIZE;


        return(TRUE);
    }

    return(FALSE);
}

static BOOL 
IMAPConnectionFreeCB(void *Buffer, void *ClientData)
{
    register ImapClient *c = (ImapClient *)Buffer;

    MemFree(c->Command);

    return(TRUE);
}

static void 
ReturnIMAPConnection(ImapClient *client)
{
    register ImapClient *c = client;
    register unsigned char  *command;

    if (c->CommandLen == BUFSIZE) {
        command = c->Command;
    } else {
        if (c->Command != NULL) {
            MemFree(c->Command);
        }

        command = (unsigned char *)MemMalloc(BUFSIZE + 1);
    }

    if (c->Command != NULL) {
        memset(c, 0, sizeof(ImapClient));

        c->state = STATE_FRESH;
        c->CommandLen = BUFSIZE;
        c->Command = command;

        c->Command[0] = '\0';

        MemPrivatePoolReturnEntry(c);
    } else {
        MemPrivatePoolDiscardEntry(c);
    }

    return;
}

BOOL 
NmapAsyncCB(void *param, char *beginPtr, char *endPtr)
{
    ImapClient *client = (ImapClient *)param;

    /* 6000 NMAP Notification of an asyncronous event */
    unsigned long event;
    unsigned char *ptr;
    unsigned long msgNum;
    unsigned long msgFlags;
    unsigned char endChar;

    endChar = *endPtr;
    *endPtr = '\0';

    ptr = beginPtr;
    event = atol(ptr);
    if (event & EVENT_MBOX_FLAGS_CHANGED) {
        event &= ~EVENT_MBOX_FLAGS_CHANGED;

        /* event msgNum msgFlags HumanReadableText */        
        ptr = strchr(ptr, ' ');
        if (ptr) {
            ptr++;
            msgNum = atol(ptr);
            ptr = strchr(ptr, ' ');
            if (ptr) {
                ptr++;
                msgFlags = atol(ptr);
                /* We need to hold onto these flags until we have a chance to tell the client about them. */

                RememberFlags(client->UnseenFlags, msgNum, msgFlags);
            }
        }
    }

    client->NMAPAsyncEvent |= event;

    *endPtr = endChar;
    return(TRUE);
}

long
LoadUIDList(ImapClient *client)
{
    unsigned long *uidList;
    unsigned long *idList;
    unsigned long count;
    unsigned long NMAPCount;
    unsigned long State;
    unsigned long UID;
    unsigned long ID;
    unsigned char Reply[1024];
    long ccode;

    if (NMAPSendCommand(client->nmap, "INFO\r\n", 6) == 6) {
        ccode = NMAPReadResponseAndCount(client->nmap, &NMAPCount);
        if (ccode == 2002) {
            uidList = MemMalloc(sizeof(unsigned long) * (NMAPCount + 1));
            if (uidList) {
                idList = MemMalloc(sizeof(unsigned long) * (NMAPCount + 1));
                if (idList) {
                    count = 0;
                    for (ID = 1; ID <= NMAPCount; ID++) {
                       ccode = ConnReadAnswer(client->nmap, Reply, sizeof(Reply));
                        if (ccode > 0) {
                            sscanf(Reply, "%*u %*u %*u %*u %lu %*s %lu", &State, &UID);
                            if (!(State & MSG_STATE_PURGED)) {
                                idList[count] = ID;
                                uidList[count] = UID;
                                count++;
                            }
                            continue;
                        }

                        MemFree(uidList);
                        MemFree(idList);
                        return(STATUS_NMAP_COMM_ERROR);
                    }

                    ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                    if (ccode == 1000) {
                        /* the lists have been successfully recreated */
                        /* free old lists and replace them with the new lists */
                        if (client->UIDList != NULL) {
                            MemFree(client->UIDList);
                        }

                        if (client->IDList != NULL) {
                            MemFree(client->IDList);
                        }

                        client->UIDList = uidList;
                        client->IDList = idList;
                        client->Messages = count;
                        return(STATUS_CONTINUE);
                    }

                    MemFree(uidList);
                    MemFree(idList);
                    return(CheckForNMAPCommError(ccode));
                }
                MemFree(uidList);
            }
            return(STATUS_MEMORY_ERROR);
        }
        return(CheckForNMAPCommError(ccode));
    }
    return(STATUS_NMAP_COMM_ERROR);
}


long
UIDtoNUM(ImapClient *client, unsigned long *UID)
{
    register unsigned long Start;
    register unsigned long End;
    register unsigned long Middle;
    BOOL tryAgain = TRUE;
    long ccode;

    if (client->Messages == 0) {
        return(STATUS_UID_NOT_FOUND);
    }

    do {
        Start = 0;
        End = client->Messages;
        Middle = End / 2;

        do {
            if (client->UIDList[Middle] > *UID) {
                End = Middle;
            } else {
                Start = Middle;
            }
            Middle = Start + (End - Start) / 2;
        } while ((End-Start) > 1);

        if (*UID == client->UIDList[Start]) {
            *UID = Start;
            return(STATUS_CONTINUE);
        }

        if (tryAgain) {
            ccode = LoadUIDList(client);
            if (ccode == STATUS_CONTINUE) {
                tryAgain = FALSE;
                continue;
            }

            return(ccode);
        }

        break;
    } while(TRUE);

    return(STATUS_UID_NOT_FOUND);
}

long
UIDtoNUMRange(ImapClient *client, unsigned long StartUID, unsigned long EndUID, long *StartIndex, long *EndIndex)
{
    register unsigned long Start;
    register unsigned long End;
    register unsigned long Middle;

    if (client->Messages==0)
        return(STATUS_UID_NOT_FOUND);

    if (StartUID>EndUID) {
        Start=EndUID;
        EndUID=StartUID;
        StartUID=Start;  
    }

    Start=0;
    End=client->Messages;
    Middle=End/2;

    do {
        if (client->UIDList[Middle]>StartUID) {
            End=Middle;
        } else {
            Start=Middle;
        }
        Middle=Start+(End-Start)/2;
    } while ((End-Start)>1);

    if (client->UIDList[Start]>=StartUID) {
        *StartIndex=Start;
    } else {
        if (Start==(client->Messages-1)) {
            if (EndUID==UID_HIGHEST) {
                *StartIndex=client->Messages-1;
                *EndIndex=*StartIndex;
                /*printf("returning special splat Start:%d End:%d\n", *StartIndex, *EndIndex); */
                return(STATUS_CONTINUE);
            }
            return(STATUS_UID_NOT_FOUND);
        } else {
            *StartIndex=++Start;
        }
    }

    if (EndUID==UID_HIGHEST) {
        *EndIndex=client->Messages-1;
        return(STATUS_CONTINUE);
    }

    /* Now check end of range */
    End=client->Messages;
    Middle=((End-Start)/2)+Start;

    do {
        if (client->UIDList[Middle]>EndUID) {
            End=Middle;
        } else {
            Start=Middle;
        }
        Middle=Start+(End-Start)/2;
    } while ((End-Start)>1);

    if (client->UIDList[Start]<=EndUID) {
        *EndIndex=Start;
    } else {
        if (Start==0) {
            return(STATUS_UID_NOT_FOUND);
        } else {
            *EndIndex=Start-1;
        }
    }

    return(STATUS_CONTINUE);
}

static BOOL
SessionCleanup(ImapClient *client)
{
    if (client) {
        if (client->state == STATE_EXITING)
            return(FALSE);
        client->state=STATE_EXITING;

        FreeUnseenFlags(client);


        if (client->nmap) {
            if (client->nmap->socket != -1) {
                NMAPSendCommand(client->nmap, "QUIT\r\n", 6);
                ConnClose(client->nmap, 1);
                client->nmap->socket = -1;
            }

            ConnFree(client->nmap);
            client->nmap = NULL;
        }
        
        if (client->conn) {
            if (client->conn->socket != -1) {
                ConnFlush(client->conn);
                ConnClose(client->conn, 1);
                client->conn->socket = -1;
            }

            ConnFree(client->conn);
            client->conn = NULL;
        }

        if (client->UIDList) {
            MemFree(client->UIDList);
        }

        if (client->IDList) {
            MemFree(client->IDList);
        }

        ReturnIMAPConnection(client);
    }
 
    XplSafeDecrement(Imap.client.threads.inUse);
    XplSafeIncrement(Imap.client.served);

    return(TRUE);
}

static long
ReadCommandLine(ImapClient *client)
{
    long ccode;
    unsigned char *tmpCommand;
    unsigned long bytesRead;

    ccode = ConnReadAnswer(client->conn, client->Command, client->CommandLen);
    if (ccode > -1) {
        if ((unsigned long)ccode < client->CommandLen) {
            return(STATUS_CONTINUE);
        } 

        bytesRead = ccode;
        do {
            tmpCommand = MemRealloc(client->Command, client->CommandLen * 2);
            if (tmpCommand) {
                client->Command = tmpCommand;
                ccode = ConnReadAnswer(client->conn, client->Command + client->CommandLen, client->CommandLen);
                if (ccode > 0) {
                    if ((unsigned long)ccode < client->CommandLen) {
                        client->CommandLen *= 2;
                        return(STATUS_CONTINUE);
                    } 

                    client->CommandLen *= 2;
                    continue;
                }
                
                if (ccode == 0) {
                    /* handle boundry condition */
                    if (client->Command[client->CommandLen - 1] == '\r') {
                        client->Command[client->CommandLen - 1] = '\0';
                    }
                    client->CommandLen *= 2;
                    return(STATUS_CONTINUE);
                }

                client->CommandLen *= 2;
                return(STATUS_ABORT);
            }

            return(STATUS_MEMORY_ERROR);
        } while (TRUE);
    }

    return(STATUS_ABORT);
}

static long
GrabOctet(ImapClient *client, unsigned char **Input, unsigned char **Destination)
{
    unsigned char *start;
    unsigned char *end;
    long Size;
    long ccode;

    /* skip '{' */
    start = *Input + 1;
    end = strchr(start, '}');

    if (end) {
        if ((end - start) < 7) { /* sanity check */
            Size = atol(start);
            if (Size) {
                *Destination = MemMalloc(Size + 1);
                if (*Destination) {
                    if ((ConnWrite(client->conn, "+ Ready for more data\r\n", 23) != -1) && (ConnFlush(client->conn) != -1)) {
                        if (ConnReadCount(client->conn, *Destination, Size) == Size) {
                            (*Destination)[Size] = '\0';

                            ccode = ReadCommandLine(client);
                            if (ccode == STATUS_CONTINUE) {
                                *Input = client->Command;
                                return(STATUS_CONTINUE);
                            }
                        }
                    }
                    MemFree(*Destination);
                    *Destination = NULL;
                    return(STATUS_ABORT);
                }
                return(STATUS_MEMORY_ERROR);
            }

            *Destination = MemStrdup("");
            if (*Destination) {
                ccode = ReadCommandLine(client);
                if (ccode == STATUS_CONTINUE) {
                    *Input = client->Command;
                    return(STATUS_CONTINUE);
                }
                MemFree(*Destination);
                *Destination = NULL;
                return(STATUS_ABORT);
            }
            return(STATUS_MEMORY_ERROR);
        }
    }
    *Destination = NULL;
    return(STATUS_INVALID_ARGUMENT);
}

static long
GrabLiteral(ImapClient *client, unsigned char **Input, unsigned char **Destination)
{
    unsigned char *contentBegin;
    unsigned long contentLen;
    unsigned char *closeQuote;

    /* skip opening quote */
    contentBegin = *Input + 1;

    if ((closeQuote = strchr(contentBegin, '"')) != NULL) {
        contentLen = closeQuote - contentBegin;
        *Destination = MemMalloc(contentLen + 1);
        if (*Destination) {
            memcpy(*Destination, contentBegin, contentLen);
            (*Destination)[contentLen] = '\0';
            *Input = closeQuote + 1;
            return(STATUS_CONTINUE);
        }

        return(STATUS_MEMORY_ERROR);
    }

    *Destination = NULL;
    return(STATUS_INVALID_ARGUMENT);
}

static unsigned char
GrabGroup(ImapClient *client, unsigned char **Input, unsigned char **Destination)
{
    unsigned char *contentBegin;
    unsigned long contentLen;
    unsigned char *closeParen;
    int nesting = 1;

    /* We have to look for the proper closing paren, since they can be nested */
    closeParen = *Input;

    do {
        closeParen++;
        if (*closeParen != ')') {
            if (*closeParen != '(') {
                if (*closeParen != '\0') {
                    continue;
                }
                *Destination = NULL;
                return(STATUS_INVALID_ARGUMENT);
            }
            nesting++;
            continue;
        }

        nesting--;
        if (nesting == 0) {
            break;
        }
    } while (TRUE);

    /* we know where the closing paren is */
    /* skip the opening paren and copy the contents */
    contentBegin = *Input + 1;
    contentLen = closeParen - contentBegin;

    *Destination = MemMalloc(contentLen + 1);
    if (*Destination) {
        memcpy(*Destination, contentBegin, contentLen);
        (*Destination)[contentLen] = '\0';
        *Input = closeParen + 1;
        return(STATUS_CONTINUE);
    }

    return(STATUS_MEMORY_ERROR);
}

long
GrabArgument(ImapClient *client, unsigned char **Input, unsigned char **Destination)
{
    unsigned char *ptr;
    unsigned char *ptr2;
    long len;

    if (*Input) {
        ptr = *Input;

        while (isspace(*ptr)) {
            ptr++;
        }

        if (*ptr == '{') {          /* octets */
            *Input = ptr;
            return(GrabOctet(client, Input, Destination));
        }
    
        if (*ptr == '"') {          /* literal */
            *Input = ptr;
            return(GrabLiteral(client, Input, Destination));
        }
    
        if (*ptr == '(') {          /* group */
            *Input = ptr;
            return(GrabGroup(client, Input, Destination));
        }

        ptr2 = strchr(ptr, ' ');
        if (ptr2) {
            len = ptr2 - ptr;
            *Destination = MemMalloc(len + 1);
            if (*Destination) {
                memcpy(*Destination, ptr, len);
                (*Destination)[len] = '\0';
                *Input = ptr2 + 1;
                return(STATUS_CONTINUE);
            }

            return(STATUS_MEMORY_ERROR);
        }

        len = strlen(ptr);
        *Destination = MemMalloc(len + 1);
        if (*Destination) {
            memcpy(*Destination, ptr, len);
            (*Destination)[len] = '\0';
            *Input = ptr + len;
            return(STATUS_CONTINUE);
        }

        return(STATUS_MEMORY_ERROR);
    }

    *Destination = MemStrdup("");
    if (*Destination) {
        return(STATUS_CONTINUE);
    }

    return(STATUS_MEMORY_ERROR);
}


static int
ConnectUserToNMAPServer(ImapClient *client, unsigned char *username, unsigned long usernameBufferSize, unsigned char *password)
{
    struct sockaddr_in nmap;
    MDBValueStruct *user;
    unsigned long len;
    unsigned char buffer[CONN_BUFSIZE + 1];

    memset(&nmap, 0, sizeof(struct sockaddr_in));

    nmap.sin_family = AF_INET;
    nmap.sin_port = htons(NMAP_PORT);

    user = MDBCreateValueStruct(Imap.directory.handle, NULL);
    if (user) {
        if (MsgFindObject(username, client->dn, NULL, &(nmap.sin_addr), user)) {
            len = strlen(user->Value[0]);

            if (len < usernameBufferSize) {
                /* aliasing in the identity store may change the name */
                strcpy(username, user->Value[0]);

                if (MsgGetUserFeature(client->dn, FEATURE_IMAP, NULL, NULL)) {
                    if (!password || MDBVerifyPassword(client->dn, password, user)) {
                        if ((client->nmap = NMAPConnectEx(NULL, &nmap, client->conn->trace.destination)) != NULL) {
                            if (NMAPAuthenticate(client->nmap, buffer, sizeof(buffer))) {
                                MDBDestroyValueStruct(user);
                                return(STATUS_CONTINUE);
                            }

                            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, nmap.sin_addr.s_addr, 0, NULL, 0);                
                            NMAPQuit(client->nmap);
                            client->nmap = NULL;
                            MDBDestroyValueStruct(user);
                            return(STATUS_NMAP_AUTH_ERROR);
                        }

                        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, nmap.sin_addr.s_addr, 0, NULL, 0);                
                        MDBDestroyValueStruct(user);
                        return(STATUS_NMAP_CONNECT_ERROR);
                    }

                    LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_WRONG_PASSWORD, LOG_NOTICE, 0, username, password, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
                    MDBDestroyValueStruct(user);
                    return(STATUS_WRONG_PASSWORD);
                }
                
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_DISABLED_FEATURE, LOG_NOTICE, 0, username, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
                MDBDestroyValueStruct(user);
                return(STATUS_FEATURE_DISABLED);
            }

            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_UNKNOWN_USER, LOG_NOTICE, 0, user->Value[0], password, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
            MDBDestroyValueStruct(user);
            return(STATUS_USERNAME_TOO_LONG);
        }

        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_UNKNOWN_USER, LOG_NOTICE, 0, username, password, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
        MDBDestroyValueStruct(user);
        return(STATUS_USERNAME_NOT_FOUND);
    }

    return(STATUS_IDENTITY_STORE_FAILURE);
}

__inline static long
PurgeMessageNotify(ImapClient *client)
{
    long ccode;
    unsigned char Answer[1024];
    unsigned char Reply[1024];
    unsigned long Exists;
    unsigned long Recent;
    unsigned long Purged;
    unsigned long newCount;
    unsigned long ExpungeCount = 0;

    if ((client->NMAPAsyncEvent & EVENT_MBOX_UNSEEN_PURGES) == 0) {
        return(STATUS_CONTINUE);
    }

    /* If we have unseen purges we want to tell the client as soon as we possibly can */
    /* Unfortunately, RFC 2180 does not permit us to report this information in response */
    /* to the STORE, FETCH or SEARCH commands.  It also does not make sense to report it */
    /* if we are already doing and Expunge */

    if (NMAPSendCommand(client->nmap, "UPDA\r\n", 6) == 6) {
        ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
        /* Any last flags should get picked up in the NmapAsyncCB() */
        SendUnseenFlags(client);

        if (ccode != 1000) {
            if (ccode == 6001) {
                do {
                    sscanf(Reply, "%lu %lu", &Exists, &Recent);
                    /*
                      Expunge messages are relative to the previous expunge, so we need to account for that 
                      in the EXPUNGE and FETCH response by reducing the msg id accordingly; this assumes
                      that NMAP returns the information in ascending order
                    */
                    ccode = NMAPtoNUM(client, Exists, &newCount);
                    if (ccode == STATUS_CONTINUE) {
                        if (Recent & MSG_STATE_PURGED) {
                            ccode = ConnWriteF(client->conn, "* %lu EXPUNGE\r\n", newCount - ExpungeCount);
                            ExpungeCount++;
                        } else {
                            long len;

                            len = sprintf(Answer, "* %lu FETCH (FLAGS (%s%s%s%s%s%s", newCount - ExpungeCount,
                                             (Recent & MSG_STATE_ANSWERED) ? "\\Answered " : "",
                                             (Recent & MSG_STATE_DELETED) ? "\\Deleted " : "",
                                             (Recent & MSG_STATE_RECENT) ? "\\Recent " : "",
                                             (Recent & MSG_STATE_READ) ? "\\Seen " : "",
                                             (Recent & MSG_STATE_DRAFT) ? "\\Draft " : "",
                                             (Recent & MSG_STATE_PRIOHIGH) ? "\\Flagged " : "");

                            if (Answer[len - 1]==' ') {
                                len--;
                            }

                            len += sprintf(Answer + len, "))\r\n");  /* This way we overwrite the last space :-) */
                            ccode = ConnWrite(client->conn, Answer, len);
                        }

                        if (ccode != -1) {
                            ;
                        } else {
                            return(STATUS_ABORT);
                        }
                    }

                    ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
                    if (ccode == 6001) {
                        continue;
                    }
                    break;
                } while(TRUE);

                if (ccode == 1000) {
                    if ((Exists != client->NMAPCount) || (ExpungeCount != 0)) {
                        sscanf(Reply, "%*s %lu %*u %lu %*u %lu", &Exists, &Recent, &Purged);

                        if (ConnWriteF(client->conn, "* %lu EXISTS\r\n* %lu RECENT\r\n", Exists - Purged, Recent) != -1) {
                            ccode = LoadUIDList(client);
                            if (ccode == STATUS_CONTINUE) {
                                client->NMAPAsyncEvent &= ~(EVENT_MBOX_NEW_MESSAGES | EVENT_MBOX_UNSEEN_PURGES);
                                return(STATUS_CONTINUE);
                            }
                            return(ccode);
                        }
                        return(STATUS_ABORT);
                    }
                    client->NMAPAsyncEvent &= ~(EVENT_MBOX_NEW_MESSAGES | EVENT_MBOX_UNSEEN_PURGES);
                    return(STATUS_CONTINUE);
                }
            }
            return(CheckForNMAPCommError(ccode));
        }
            
        client->NMAPAsyncEvent &= ~(EVENT_MBOX_NEW_MESSAGES | EVENT_MBOX_UNSEEN_PURGES);
        return(STATUS_CONTINUE);
    }
    return(STATUS_NMAP_COMM_ERROR);
}

long
NewMessageNotify(ImapClient *client)
{
    long ccode;
    unsigned char Reply[1024];
    unsigned long Exists;
    unsigned long Recent;
    unsigned long Purged;


    if ((client->NMAPAsyncEvent & EVENT_MBOX_NEW_MESSAGES) == 0) {
        return(STATUS_CONTINUE);
    }

    /* If we have new messages, we want to tell the client as soon as we possibly can.  */
    /* Untagged EXIST and RECENT responses is how this is done.  If the current command    */
    /* changes the mailbox state, (Close, Logout, Select) we don't want to confuse the client with these responses.*/
    /* They are redundant on the EXPUNGE command since it sends EXIST and RECENT already. */


    if (NMAPSendCommand(client->nmap, "STAT\r\n", 6) == 6) {
        ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
        if (ccode == 1000) {
            sscanf(Reply, "%*s %lu %*u %lu %*u %lu", &Exists, &Recent, &Purged);

            if (Exists > client->NMAPCount) {
                ccode = LoadUIDList(client);
                if (ccode == STATUS_CONTINUE) {
                    client->NMAPAsyncEvent &= ~EVENT_MBOX_NEW_MESSAGES;
                    if (ConnWriteF(client->conn, "* %lu EXISTS\r\n* %lu RECENT\r\n", Exists - Purged, Recent) != -1) {
                        return(STATUS_CONTINUE);
                    }
                    return(STATUS_ABORT);
                }
        
                return(ccode);
            }

            client->NMAPAsyncEvent &= ~EVENT_MBOX_NEW_MESSAGES;
            if (ConnWriteF(client->conn, "* %lu EXISTS\r\n* %lu RECENT\r\n", Exists - Purged, Recent) != -1) {
                return(STATUS_CONTINUE);
            }

            return(STATUS_ABORT);
        }
        return(CheckForNMAPCommError(ccode));
    }
    return(STATUS_NMAP_COMM_ERROR);
}

static int
ImapCommandUnknown(void *param)
{
    ImapClient *client = (ImapClient *)param;

    LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_UNHANDLED, LOGGER_EVENT_UNHANDLED_REQUEST, LOG_INFO, 0, 
                client->User, 
                client->Command, 
                XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 
                0, 
                NULL, 
                0);
    
    return(SendError(client->conn, "*", client->Ident, STATUS_UNKNOWN_COMMAND));
}

/********** IMAP client commands - any state (Capability, Noop, Logout) **********/
int
ImapCommandCapability(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "CAPABILITY", ccode));
        }
    }

    if (!Imap.server.ssl.enable || client->conn->ssl.enable) {
        ccode = ConnWrite(client->conn, Imap.command.capability.message, Imap.command.capability.len);
    } else {
        ccode = ConnWrite(client->conn, Imap.command.capability.ssl.message, Imap.command.capability.ssl.len);
    }

    return(SendOk(client, "CAPABILITY"));
}

int
ImapCommandNoop(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;

    if (client->state == STATE_SELECTED) {
        if (NMAPSendCommand(client->nmap, "NOOP\r\n", 6) == 6) {
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                    return(SendOk(client, "NOOP"));
                }
                return(SendError(client->conn, client->Ident, "NOOP", ccode));
            }
            return(SendError(client->conn, client->Ident, "NOOP", CheckForNMAPCommError(ccode)));
        }
        return(SendError(client->conn, client->Ident, "NOOP", STATUS_NMAP_COMM_ERROR));
    }
    return(SendOk(client, "NOOP"));
}

int
ImapCommandLogout(void *param)
{
    ImapClient *client = (ImapClient *)param;

    if (ConnWrite(client->conn, "* BYE IMAP4rev1 Server signing off\r\n", 36) != -1) {
        if (SendOk(client,"LOGOUT") == STATUS_CONTINUE) {
            ConnFlush(client->conn);
        }
    }
    
    return(STATUS_ABORT);
}


/********** IMAP client commands - not authenticated state (STARTTLS, AUTHENTICATE, LOGIN) **********/

int
ImapCommandStartTls(void *param)
{
    long ccode;

    ImapClient *client = (ImapClient *)param;

    if (client->state < STATE_AUTH) {
        if (Imap.server.ssl.enable) {
            if (((ccode = ConnWriteF(client->conn, "%s OK Begin TLS negotiation\r\n",client->Ident)) != -1) && (ccode = ConnFlush(client->conn))) {
                client->conn->ssl.enable = TRUE;    
                if (ConnNegotiate(client->conn, Imap.server.ssl.context)) {
                    return(STATUS_CONTINUE);
                }

                return(SendError(client->conn, client->Ident, "STARTTLS", STATUS_TLS_NEGOTIATION_FAILURE));
            } 

            return(STATUS_ABORT);
        }

        return(SendError(client->conn, client->Ident, "", STATUS_UNKNOWN_COMMAND));
    }

    return(SendError(client->conn, client->Ident, "STARTTLS", STATUS_INVALID_STATE));


}

int
ImapCommandAuthenticate(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char Answer[1024];
    unsigned char Reply[1024];

    unsigned char *PW;
    unsigned long len;

    if (client->state == STATE_FRESH) {
        if (XplStrNCaseCmp(client->Command+13,"LOGIN", 5) == 0) {
            if ((ccode = ConnWrite(client->conn, "+ VXNlcm5hbWU6AA==\r\n", 20) != -1) && (ccode = ConnFlush(client->conn)) != -1) {
                ccode = ReadCommandLine(client);
                if (ccode == STATUS_CONTINUE) {
                    len = strlen(client->Command);
                    if (len < sizeof(Reply)) {
                        memcpy(Reply, client->Command, len);
                        Reply[len] = '\0';

                        if ((ConnWrite(client->conn, "+ UGFzc3dvcmQ6AA==\r\n", 20) != -1) && (ConnFlush(client->conn) != -1)) {
                            ccode = ReadCommandLine(client);
                            if (ccode == STATUS_CONTINUE) {
                                PW=DecodeBase64(client->Command);
                                DecodeBase64(Reply);

                                len = strlen(Reply);
                                if (len < sizeof(client->User)) {
                                    strcpy(client->User, Reply);

                                    ccode = ConnectUserToNMAPServer(client, client->User, sizeof(client->User) - 1,  PW);
                                    if (ccode == STATUS_CONTINUE) {
                                        CMAuthenticated((unsigned long)client->conn->socketAddress.sin_addr.s_addr, client->User);

                                        /* Set up NMAP event handler */
                                        client->nmap->client.cb = (void *)NmapAsyncCB;
                                        client->nmap->client.data = (void *)client;

                                        /* Request NMAP events */
                                        if (NMAPSendCommandF(client->nmap, "FLAG %lu\r\n",(unsigned long)(NMAP_SHOWDELETED | NMAP_OOBMESSAGES)) != -1) {
                                            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                                            if (ccode == 1000) {
                                                if (NMAPSendCommandF(client->nmap, "USER %s\r\n",client->User) != -1) {
                                                    ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                                                    if (ccode == 1000) {
                                                        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_LOGIN, LOG_INFO, 0, client->User, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
                                                        snprintf(Answer, sizeof(Answer), "IMAP:%s", client->User);
                                                        XplRenameThread(XplGetThreadID(), Answer);
                                                        client->state = STATE_AUTH;
                                                        return(SendOk(client, "AUTHENTICATE"));
                                                    }
                                                    LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                                    return(SendError(client->conn, client->Ident, "AUTHENTICATE", CheckForNMAPCommError(ccode)));
                                                }
                                                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                                return(SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_NMAP_COMM_ERROR));
                                            }
                                            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                            return(SendError(client->conn, client->Ident, "AUTHENTICATE", CheckForNMAPCommError(ccode)));
                                        }
                                        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                        return(SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_NMAP_COMM_ERROR));
                                    } 

                                    if ((ccode == STATUS_WRONG_PASSWORD) || (ccode == STATUS_USERNAME_NOT_FOUND)) {
                                        XplDelay(2000);
                                        client->LoginCount++;
                                        if (client->LoginCount >= MAX_FAILED_LOGINS) {
                                            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_MAX_FAILED_LOGINS, LOG_NOTICE, 0, client->User, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), MAX_FAILED_LOGINS, NULL, 0);
                                            if (SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_PERMISSION_DENIED) != STATUS_ABORT) {
                                                if (ConnWrite(client->conn, "* BYE IMAP4rev1 Server signing off\r\n", 36) != -1) {
                                                    ConnFlush(client->conn);
                                                }
                                            }
                                            return(STATUS_ABORT);
                                        }

                                        return(SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_PERMISSION_DENIED));
                                    }

                                    return(SendError(client->conn, client->Ident, "AUTHENTICATE", ccode));
                                }
                                if (ConnWrite(client->conn, "* BYE IMAP4rev1 Server signing off\r\n", 36) != -1) {
                                    SendError(client->conn, client->Ident, "Authenticate", STATUS_INVALID_ARGUMENT);
                                }
                            }
                        }
                        return(STATUS_ABORT);
                    }

                    if (ConnWrite(client->conn, "* BYE IMAP4rev1 Server signing off\r\n", 36) != -1) {
                        SendError(client->conn, client->Ident, "Authenticate", STATUS_INVALID_ARGUMENT);
                    }
                    return(STATUS_ABORT);
                }
                return(SendError(client->conn, client->Ident, "AUTHENTICATE", ccode));
            }
            return(STATUS_ABORT);
        }
        return(SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_UNKNOWN_AUTH));
    }
    return(SendError(client->conn, client->Ident, "AUTHENTICATE", STATUS_INVALID_STATE));
}

int
ImapCommandLogin(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char *spacePtr;
    unsigned char Answer[1024];
    unsigned char *username = NULL;
    unsigned char *password = NULL;

    if (client->state == STATE_FRESH) {
        ptr = client->Command + 6;
        while (isspace(*ptr)) {
            ptr++;
        }

        /* ptr points to username */
        if (*ptr) {
            ccode = GrabArgument(client, &ptr, &username);
            if (ccode == STATUS_CONTINUE) {
                if (ptr) {
                    while ((spacePtr = strchr(username, ' ')) != NULL) {
                        *spacePtr = '_';
                    }

                    if (strlen(username) < sizeof(client->User)) {
                        strcpy(client->User, username);
                        MemFree(username);
                        username = NULL;

                        ccode = GrabArgument(client, &ptr, &password);
                        if (ccode == STATUS_CONTINUE) {
                            ccode = ConnectUserToNMAPServer(client, client->User, sizeof(client->User) - 1, password);
                            MemFree(password);

                            switch (ccode) {
                                case STATUS_CONTINUE: {
                                    client->state = STATE_AUTH;
                                    break;
                                }

                                case STATUS_WRONG_PASSWORD:
                                case STATUS_USERNAME_NOT_FOUND: {
                                    XplDelay(2000);
                                    client->LoginCount++;
                                    if (client->LoginCount >= MAX_FAILED_LOGINS) {
                                        if (ConnWrite(client->conn, "* BYE IMAP4rev1 Server signing off\r\n", sizeof("* BYE IMAP4rev1 Server signing off\r\n") - 1) != -1) {
                                            if (SendError(client->conn, client->Ident, "LOGIN", STATUS_PERMISSION_DENIED) == STATUS_CONTINUE) {
                                                ConnFlush(client->conn);    
                                            }
                                        }
                                        
                                        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_MAX_FAILED_LOGINS, LOG_NOTICE, 0, client->User, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), MAX_FAILED_LOGINS, NULL, 0);
                                        return(STATUS_ABORT);
                                    }

                                    return(SendError(client->conn, client->Ident, "LOGIN", STATUS_PERMISSION_DENIED));
                                }

                                default: {
                                    return(SendError(client->conn, client->Ident, "LOGIN", ccode));
                                }
                            }

                            CMAuthenticated((unsigned long)client->conn->socketAddress.sin_addr.s_addr, client->User);
                            /* Set up NMAP event handler */
                            client->nmap->client.cb = (void *)NmapAsyncCB;
                            client->nmap->client.data = (void *)client;

                            /* Request NMAP events */
                            if (NMAPSendCommandF(client->nmap, "FLAG %lu\r\n",(unsigned long)(NMAP_SHOWDELETED | NMAP_OOBMESSAGES)) != -1) {
                                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                                if (ccode == 1000) {
                                    if (NMAPSendCommandF(client->nmap, "USER %s\r\n",client->User) != -1) {
                                        ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                                        if (ccode == 1000) {
                                            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_LOGIN, LOG_INFO, 0, username, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
                                            snprintf(Answer, sizeof(Answer), "IMAP:%s", client->User);
                                            XplRenameThread(XplGetThreadID(), Answer);
                                            client->state = STATE_AUTH;
                                            return(SendOk(client,"LOGIN"));
                                        }
                                        
                                        return(SendError(client->conn, client->Ident, "LOGIN", CheckForNMAPCommError(ccode)));
                                    }
                                    LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                    return(SendError(client->conn, client->Ident, "LOGIN", STATUS_NMAP_COMM_ERROR));
                                }
                                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                                return(SendError(client->conn, client->Ident, "LOGIN", CheckForNMAPCommError(ccode)));
                            }
                            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_NMAP_UNAVAILABLE, LOG_ERROR, 0, NULL, NULL, client->nmap->socketAddress.sin_addr.s_addr, ccode, NULL, 0);
                            return(SendError(client->conn, client->Ident, "LOGIN", STATUS_NMAP_COMM_ERROR));
                        }
                    }
                }
                MemFree(username);
                return(SendError(client->conn, client->Ident, "LOGIN", STATUS_INVALID_ARGUMENT));
            }
            return(SendError(client->conn, client->Ident, "LOGIN", ccode));
        }
        return(SendError(client->conn, client->Ident, "LOGIN", STATUS_INVALID_ARGUMENT));
    }
    return(SendError(client->conn, client->Ident, "LOGIN", STATUS_INVALID_STATE));
}

/********** IMAP client commands - authenticated state **********/

int
ImapCommandSelect(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];

    unsigned char *mbox;
    unsigned long Exists;
    unsigned long Recent;
    unsigned long Purged;
    unsigned long BoxID;
    BOOL IsExamine = FALSE;

    if (client->state == STATE_SELECTED) {
        if (NMAPSendCommand(client->nmap, "RSET MBOX\r\n", 11) == 11) {
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);

            /* Any events are for the old mailbox, so forget about them. */
            client->NMAPAsyncEvent = 0;

            client->state = STATE_AUTH;

            if (ccode == 1000) {
                ;
            } else {
                return(SendError(client->conn, client->Ident, "SELECT", CheckForNMAPCommError(ccode)));
            }
        } else {
            return(SendError(client->conn, client->Ident, "SELECT", STATUS_NMAP_COMM_ERROR));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "SELECT", STATUS_INVALID_STATE));
    }

    ptr = client->Command + 7;
    if (toupper(*(ptr-1)) == 'E') {
        IsExamine = TRUE;
    } else if (*(ptr-1) != ' ') {
        return(SendError(client->conn, client->Ident, "SELECT", STATUS_INVALID_ARGUMENT));
    }

    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);
        CheckPathLen(mbox);

        if (strlen(mbox) < sizeof(client->MBox)) {
            strcpy(client->MBox, mbox);
            MemFree(mbox);
        } else {
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "SELECT", STATUS_INVALID_FOLDER_NAME));
        }

        if((client->MBox[0] == '/') || (client->MBox[0] == '\\')) {
            ccode = NMAPSendCommandF(client->nmap, "MBOX %s\r\n", client->MBox + 1);
        } else {
            ccode = NMAPSendCommandF(client->nmap, "MBOX %s\r\n", client->MBox);
        }

        if (ccode != -1) {
            ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
            if (ccode != NMAP_MBOX_RW_SELECTED) {
                if (ccode != NMAP_MBOX_RO_SELECTED) {
                    return(SendError(client->conn, client->Ident, "SELECT", CheckForNMAPCommError(ccode)));
                } else {
                    client->ReadOnly=TRUE;
                }
            }
            BoxID = atol(Reply);

            if (NMAPSendCommand(client->nmap, "STAT\r\n", 6) == 6) {
                ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
                if (ccode == 1000) {
                    sscanf(Reply, "%*s %lu %*u %lu %*u %lu", &Exists, &Recent, &Purged);
                    if (ConnWriteF(client->conn, "* %lu EXISTS\r\n* %lu RECENT\r\n", Exists - Purged, Recent) != -1) {
                        ccode = LoadUIDList(client);
                        if (ccode == STATUS_CONTINUE) {
                            if (ConnWriteF(client->conn, "* FLAGS (\\Answered \\Flagged \\Deleted \\Draft \\Seen)\r\n* OK [UIDVALIDITY %lu] UID validity status\r\n",BoxID) != -1) {
                                if (IsExamine || client->ReadOnly) {
                                    if (ConnWrite(client->conn, "* OK [PERMANENTFLAGS ()] Permanent flags\r\n", 42) != -1) {
                                        if (ConnWriteF(client->conn, "%s OK [READ-ONLY] %s completed\r\n",client->Ident, IsExamine ? "EXAMINE" : "SELECT") != -1) {
                                            client->state=STATE_SELECTED;
                                            return(STATUS_CONTINUE);
                                        }
                                    }
                                    return(STATUS_ABORT);
                                }

                                if (ConnWrite(client->conn, "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Draft \\Seen)] Permanent flags\r\n", 82) != -1) {
                                    if (ConnWriteF(client->conn, "%s OK [READ-WRITE] SELECT completed\r\n",client->Ident) != -1) {
                                        client->state=STATE_SELECTED;
                                        return(STATUS_CONTINUE);
                                    }
                                }
                            }
                            return(STATUS_ABORT);
                        }
                        return(SendError(client->conn, client->Ident, "SELECT", ccode));
                    }
                    return(STATUS_ABORT);
                }
                return(SendError(client->conn, client->Ident, "SELECT", CheckForNMAPCommError(ccode)));
            }
        }
        return(SendError(client->conn, client->Ident, "SELECT", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "SELECT", ccode));
}

int
ImapCommandExamine(void *param)
{
    return(ImapCommandSelect(param));
}

int
ImapCommandCreate(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    unsigned char *mbox;
    int len;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "CREATE", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "CREATE", STATUS_INVALID_STATE));
    }

    ptr=client->Command+2;
    while(!isspace(*ptr) && *ptr != '\0')
        ++ptr;
    if(*ptr == '\0') {
        return(SendError(client->conn, client->Ident, "CREATE", STATUS_INVALID_ARGUMENT));
    }

    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        if (mbox[0]=='\0') {
            return(SendError(client->conn, client->Ident, "CREATE", STATUS_INVALID_ARGUMENT));
        }
        RemoveMboxSpaces(mbox);
        CheckPathLen(mbox);
        len = strlen(mbox);

        if (XplStrCaseCmp(mbox,client->User) == 0) {
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "CREATE", STATUS_INVALID_FOLDER_NAME));
        }
        while(len > 0 && (mbox[len - 1] == '/' || mbox[len - 1] == '\\')) {
            mbox[--len] = '\0';
        }
        if (NMAPSendCommandF(client->nmap, "CREA %s\r\n",mbox) != -1) {
            MemFree(mbox);
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                return(SendOk(client,"CREATE"));
            }
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "CREATE", CheckForNMAPCommError(ccode)));
        }
        MemFree(mbox);
        return(SendError(client->conn, client->Ident, "CREATE", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "CREATE", ccode));
}

int
ImapCommandDelete(void *param)
{
    long ccode;
    long reason;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    BOOL repeat = FALSE;
    unsigned char *mbox;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "DELETE", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "DELETE", STATUS_INVALID_STATE));
    }

    ptr = client->Command + 7;
    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);
        CheckPathLen(mbox);
        if(XplStrCaseCmp(mbox, "INBOX") != 0) {
            do {
                if (NMAPSendCommandF(client->nmap, "RMOV %s\r\n",mbox) != -1) {
                    reason = NMAPReadResponse(client->nmap, NULL, 0, 0);
                    if (reason == 1000) {
                        return(SendOk(client, "DELETE"));
                    }

                    if(reason == 4227 && client->state == STATE_SELECTED) {
                        repeat = repeat ? FALSE : TRUE; /* Only try once */
                        if (NMAPSendCommand(client->nmap, "RSET MBOX\r\n", 11) == 11) {
                            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                            if(ccode == 1000) {
                                client->state = STATE_AUTH;
                                client->MBox[0] = '\0';
                                continue;
                            }
                            MemFree(mbox);
                            return(SendError(client->conn, client->Ident, "DELETE", CheckForNMAPCommError(ccode)));
                        }
                        MemFree(mbox);
                        return(SendError(client->conn, client->Ident, "DELETE", STATUS_NMAP_COMM_ERROR));
                    }
                    MemFree(mbox);
                    return(SendError(client->conn, client->Ident, "DELETE", CheckForNMAPCommError(reason)));
                }
                MemFree(mbox);
                return(SendError(client->conn, client->Ident, "DELETE", STATUS_NMAP_COMM_ERROR));
            } while(repeat);
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "DELETE", CheckForNMAPCommError(reason)));
        }
        MemFree(mbox);
        return(SendError(client->conn, client->Ident, "DELETE", STATUS_INVALID_FOLDER_NAME));
    }
    return(SendError(client->conn, client->Ident, "DELETE", ccode));
}

int
ImapCommandRename(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    unsigned char *newbox;
    unsigned char *mbox;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "RENAME", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "RENAME", STATUS_INVALID_STATE));
    }

    ptr=client->Command+7;
    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);
        ccode = GrabArgument(client, &ptr, &newbox);
        if (ccode == STATUS_CONTINUE) {
            RemoveMboxSpaces(newbox);
            if ((client->state == STATE_SELECTED) && (XplStrCaseCmp(client->MBox, mbox) == 0)) { 
                if (NMAPSendCommand(client->nmap, "RSET MBOX\r\n", 11) == 11) {
                    ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                    if (ccode == 1000) {
                        ;
                    } else {
                        MemFree(newbox);
                        MemFree(mbox);
                        return(SendError(client->conn, client->Ident, "RENAME", CheckForNMAPCommError(ccode)));
                    }
                } else {
                    MemFree(newbox);
                    MemFree(mbox);
                    return(SendError(client->conn, client->Ident, "RENAME", STATUS_NMAP_COMM_ERROR));
                }
            }

            CheckPathLen(mbox);
            CheckPathLen(newbox);

            if (NMAPSendCommandF(client->nmap, "RNAM %s %s\r\n",mbox, newbox) != -1) {
                MemFree(newbox);
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode == 1000) {
                    if (XplStrCaseCmp(mbox, "INBOX") == 0) {
                        if (NMAPSendCommand(client->nmap, "CREA INBOX\r\n", 12) == 12) {
                            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                        } else {
                            MemFree(mbox);
                            return(SendError(client->conn, client->Ident, "RENAME", STATUS_NMAP_COMM_ERROR));
                        }
                    
                        if (ccode == 1000) {
                            ;
                        } else {
                            MemFree(mbox);
                            return(SendError(client->conn, client->Ident, "RENAME", CheckForNMAPCommError(ccode)));
                        }
                    }

                    if (XplStrCaseCmp(client->MBox, mbox) == 0) {
                        MemFree(mbox);
                        if (NMAPSendCommandF(client->nmap, "MBOX %s\r\n",client->MBox) != -1) {
                            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                            if (ccode == 1000) {
                                ccode = LoadUIDList(client);
                                if (ccode == STATUS_CONTINUE) {
                                    return(SendOk(client, "RENAME"));
                                }
                                return(SendError(client->conn, client->Ident, "RENAME", ccode));
                            }
                            return(SendError(client->conn, client->Ident, "RENAME", CheckForNMAPCommError(ccode)));
                        }
                        return(SendError(client->conn, client->Ident, "RENAME", STATUS_NMAP_COMM_ERROR));
                    }

                    MemFree(mbox);
                    return(SendOk(client, "RENAME"));
                }
                MemFree(mbox);
                return(SendError(client->conn, client->Ident, "RENAME", CheckForNMAPCommError(ccode)));
            }
            MemFree(newbox);
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "RENAME", STATUS_NMAP_COMM_ERROR));
        }
        MemFree(mbox);
    }
    return(SendError(client->conn, client->Ident, "RENAME", ccode));
}

int
ImapCommandSubscribe(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    unsigned char *mbox;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "SUBSCRIBE", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "SUBSCRIBE", STATUS_INVALID_STATE));
    }

    ptr = client->Command + 9;
    while(isspace(*ptr)) {
        ptr++;
    }

    if (!*ptr) {
        return(SendError(client->conn, client->Ident, "SUBSCRIBE", STATUS_INVALID_ARGUMENT));
    }
       
    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);

        if (NMAPSendCommandF(client->nmap, "SUBS %s\r\n", mbox) != -1) {
            MemFree(mbox);
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                return(SendOk(client, "SUBSCRIBE"));
            }
            return(SendError(client->conn, client->Ident, "SUBSCRIBE", CheckForNMAPCommError(ccode)));
        }
        return(SendError(client->conn, client->Ident, "SUBSCRIBE", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "SUBSCRIBE", ccode));
}

int
ImapCommandUnsubscribe(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    unsigned char *mbox;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", STATUS_INVALID_STATE));
    }

    ptr = client->Command + 11;
    while(isspace(*ptr)) {
        ptr++;
    }

    if (!*ptr) {
        return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", STATUS_INVALID_ARGUMENT));
    }
       
    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);

        if (NMAPSendCommandF(client->nmap, "UNSUBS %s\r\n", mbox) != -1) {
            MemFree(mbox);
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                return(SendOk(client, "UNSUBSCRIBE"));
            }
            return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", CheckForNMAPCommError(ccode)));
        }
        MemFree(mbox);
        return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "UNSUBSCRIBE", ccode));
}

int
ImapCommandList(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;
    unsigned char Reply[1024];

    unsigned char *pattern;
    int    len;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "LIST", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "LIST", STATUS_INVALID_STATE));
    }

    ptr=client->Command+5;
    while(isspace(*ptr))
        ptr++;

    if (!*ptr) {
        return(SendError(client->conn, client->Ident, "LIST", STATUS_INVALID_ARGUMENT));
    }
         
    ccode = GrabArgument(client, &ptr, &pattern);
    if (ccode == STATUS_CONTINUE) {
        MemFree(pattern);
        ccode = GrabArgument(client, &ptr, &pattern);
        if (ccode == STATUS_CONTINUE) {
            RemoveMboxSpaces(pattern);
            len = strlen(pattern);
            while(len > 0 && (pattern[len - 1] == '/' || pattern[len - 1] == '\\')) {
                pattern[--len] = '\0';
            }

            /* Handle RFC special command "" on the pattern */
            if (pattern[0]=='\0') {
                MemFree(pattern);
                if (ConnWrite(client->conn, "* LIST (\\Noselect) \"/\" \"\"\r\n", sizeof("* LIST (\\Noselect) \"/\" \"\"\r\n") - 1) != -1) {
                    return(SendOk(client, "LIST"));
                }
                return(STATUS_ABORT);
                
            }


            if (NMAPSendCommandF(client->nmap, "SHOW %s\r\n",pattern) != -1) {
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode != 2002) {
                    MemFree(pattern);
                    if (ccode == 1000) {
                        return(SendError(client->conn, client->Ident, "LIST", STATUS_NO_MATCHING_FOLDER));
                    }

                    return(SendError(client->conn, client->Ident, "LIST", CheckForNMAPCommError(ccode)));
                }

                while ((ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE)) == 2002 && !Imap.exiting) {
                    AddMboxSpaces(Reply);
                    if (Reply[strlen(Reply)-1]!='/') { 
                        ccode = ConnWriteF(client->conn, "* LIST () \"/\" \"%s\"\r\n",Reply);
                    } else {
                        Reply[strlen(Reply)-1]='\0';
                        ccode = ConnWriteF(client->conn, "* LIST (\\NoSelect) \"/\" \"%s\"\r\n",Reply);
                    }
                    if (ccode != -1) {
                        continue;
                    }
                    MemFree(pattern);
                    return(STATUS_ABORT);
                }
                if (ccode == 1000) {
                    MemFree(pattern);
                    return(SendOk(client, "LIST"));
                }
                return(SendError(client->conn, client->Ident, "LIST", CheckForNMAPCommError(ccode)));
            }
            return(SendError(client->conn, client->Ident, "LIST", STATUS_NMAP_COMM_ERROR));
        }
    }

    return(SendError(client->conn, client->Ident, "LIST", ccode));
}

int
ImapCommandLsub(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];
    unsigned char *pattern;
    int    len;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "LSUB", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "LSUB", STATUS_INVALID_STATE));
    }

    ptr=client->Command+5;
    while(isspace(*ptr))
        ptr++;

    if (!*ptr) {
        return(SendError(client->conn, client->Ident, "LSUB", STATUS_INVALID_ARGUMENT));
    }
       
    ccode = GrabArgument(client, &ptr, &pattern);
    if (ccode == STATUS_CONTINUE) {
        MemFree(pattern);
        ccode = GrabArgument(client, &ptr, &pattern);
        if (ccode == STATUS_CONTINUE) {
            RemoveMboxSpaces(pattern);
            len = strlen(pattern);
            while(len > 0 && (pattern[len - 1] == '/' || pattern[len - 1] == '\\')) {
                pattern[--len] = '\0';
            }

            /* Handle RFC special command "" on the pattern */
            if (pattern[0]=='\0') {
                MemFree(pattern);
                if (ConnWrite(client->conn, "* LSUB (\\Noselect) \"/\" \"\"\r\n", sizeof("* LSUB (\\Noselect) \"/\" \"\"\r\n") - 1) != -1) {
                    return(SendOk(client, "LSUB"));
                }
                return(STATUS_ABORT);
            }


            if (NMAPSendCommandF(client->nmap, "SHOWSUB %s\r\n",pattern) != -1) {
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode != 2002) {
                    MemFree(pattern);
                    if (ccode == 1000) {
                        return(SendError(client->conn, client->Ident, "LSUB", STATUS_NO_MATCHING_FOLDER));
                    }
                    return(SendError(client->conn, client->Ident, "LSUB", CheckForNMAPCommError(ccode)));
                }

                while ((ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE)) == 2002 && !Imap.exiting) {
                    AddMboxSpaces(Reply);
                    if (Reply[strlen(Reply)-1]!='/') { 
                        ccode = ConnWriteF(client->conn, "* LSUB () \"/\" \"%s\"\r\n",Reply);
                    } else {
                        Reply[strlen(Reply)-1]='\0';
                        ccode = ConnWriteF(client->conn, "* LSUB (\\NoSelect) \"/\" \"%s\"\r\n", Reply);
                    }
                    if (ccode != -1) {
                        continue;
                    }
                    MemFree(pattern);
                    return(STATUS_ABORT);
                }
                if (ccode == 1000) {
                    MemFree(pattern);
                    return(SendOk(client, "LSUB"));
                }
                return(SendError(client->conn, client->Ident, "LSUB", CheckForNMAPCommError(ccode)));
            }
            MemFree(pattern);
            return(SendError(client->conn, client->Ident, "LSUB", STATUS_NMAP_COMM_ERROR));
        }
    }
    return(SendError(client->conn, client->Ident, "LSUB", ccode));
}

typedef struct {
    unsigned long totalCount;
    unsigned long recentCount;
    unsigned long readCount;
    unsigned long purgedCount;
    unsigned long uidNext;
    unsigned long mboxUid;
} CheckResult;

static unsigned long
CheckMessages(CheckResult *check)
{
    return(check->totalCount - check->purgedCount);
}

static unsigned long
CheckRecent(CheckResult *check)
{
    return(check->recentCount);
}

static unsigned long
CheckUidnext(CheckResult *check)
{
    return(check->uidNext);
}

static unsigned long
CheckUidvalidity(CheckResult *check)
{
    return(check->mboxUid);
}

static unsigned long
CheckUnseen(CheckResult *check)
{
    return(check->totalCount - check->purgedCount - check->readCount);
}

typedef unsigned long (* CheckResultFunction)(CheckResult *check);

typedef struct {
    unsigned char *string;
    unsigned long stringLen;
    CheckResultFunction function;
} StatusCommandOption;

static StatusCommandOption StatusCommandOptions[] = {
    { "MESSAGES",       sizeof("MESSAGES") - 1,     CheckMessages },
    { "RECENT",         sizeof("RECENT") - 1,       CheckRecent },
    { "UIDNEXT",        sizeof("UIDNEXT") - 1,      CheckUidnext },
    { "UIDVALIDITY",    sizeof("UIDVALIDITY") - 1,  CheckUidvalidity },
    { "UNSEEN",         sizeof("UNSEEN") - 1,       CheckUnseen },
    { NULL,             0,                          0 }
};

BOOL
CommandStatusInit()
{
    HulaKeywordIndexCreateFromTable(Imap.command.status.optionIndex, StatusCommandOptions, .string, TRUE);
    if (Imap.command.status.optionIndex) {
        return(TRUE);
    }
    return(FALSE);
}

void
CommandStatusCleanup()
{
    HulaKeywordIndexFree(Imap.command.status.optionIndex);
}

int
ImapCommandStatus(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];
    unsigned char *mbox;
    unsigned char *items;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "STATUS", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "STATUS", STATUS_INVALID_STATE));
    }

    ptr=client->Command+6;
    while(isspace(*ptr))
        ptr++;

    if (!*ptr) {
        return(SendError(client->conn, client->Ident, "STATUS", STATUS_INVALID_ARGUMENT));
    }
         
    ccode = GrabArgument(client, &ptr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);

        ccode = GrabArgument(client, &ptr, &items);
        if (ccode == STATUS_CONTINUE) {
            if (NMAPSendCommandF(client->nmap, "CHECK %s\r\n", mbox) != -1) {
                ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
                if (ccode == 1000) {
                    CheckResult checkResults;
                    unsigned char *ptr,*DoSpace="";

                    ptr=items;
                    if (*ptr=='(')
                        ptr++;

                    // Total, Recent, Purged, Read, IDs
                    sscanf(Reply, "%lu %*u %lu %*u %lu %*u %lu %*u %lu %lu", 
                                    &checkResults.totalCount, 
                                    &checkResults.recentCount, 
                                    &checkResults.purgedCount, 
                                    &checkResults.readCount, 
                                    &checkResults.uidNext, 
                                    &checkResults.mboxUid);

                    AddMboxSpaces(mbox);
                    if (ConnWriteF(client->conn, "* STATUS \"%s\" (", mbox) != -1) {
                        long option;
                        while (*ptr) {
                            option = HulaKeywordFind(Imap.command.status.optionIndex, ptr);
                            if (option != -1) {
                                if (ConnWriteF(client->conn, "%s%s %lu", DoSpace, StatusCommandOptions[option].string, StatusCommandOptions[option].function(&checkResults)) != -1) {
                                    ptr += StatusCommandOptions[option].stringLen;
                                    if (*ptr == ' ') {
                                        ptr++;
                                        DoSpace = " ";
                                        continue;
                                    }
                                    break;
                                }
                                MemFree(items);
                                MemFree(mbox);
                                return(STATUS_ABORT);
                            }
                            MemFree(items);
                            MemFree(mbox);
                            if (ConnWrite(client->conn, "\r\n", sizeof("\r\n") - 1) != -1) {
                                return(SendError(client->conn, client->Ident, "STATUS", STATUS_INVALID_ARGUMENT));
                            }
                            return(STATUS_ABORT);
                        }
                    
                        MemFree(items);
                        MemFree(mbox);
                        if (ConnWrite(client->conn, ")\r\n", 3) != -1) {
                            return(SendOk(client, "STATUS"));
                        }
                        return(STATUS_ABORT);
                    }
                    MemFree(items);
                    MemFree(mbox);
                    return(STATUS_ABORT);
                }
                MemFree(items);
                MemFree(mbox);
                return(SendError(client->conn, client->Ident, "STATUS", CheckForNMAPCommError(ccode)));
            }
            MemFree(mbox);
            MemFree(items);
            return(SendError(client->conn, client->Ident, "STATUS", STATUS_NMAP_COMM_ERROR));
        }
        MemFree(mbox);
    }
    return(SendError(client->conn, client->Ident, "STATUS", ccode));
}

int
ImapCommandAppend(void *param)
{
    ImapClient *client = (ImapClient *)param;
    unsigned char *mbox=NULL, *tmp, *ptr2;
    unsigned long Flags=0;
    long Size;
    unsigned long Date=0;
    long ccode;
    unsigned char *ptr;
    unsigned char *argPtr;

    if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "APPEND", STATUS_INVALID_STATE));
    }

    ptr=client->Command+6;
    if (*ptr!=' ') {
        return(SendError(client->conn, client->Ident, "APPEND", STATUS_INVALID_ARGUMENT));
    }

    while (isspace(*ptr)) {
        ptr++;
    }

    /* since ptr gets used below, we need to preserve it */
    /* because GrabArgument advances the pointer passed in */
    argPtr = ptr;
    ccode = GrabArgument(client, &argPtr, &mbox);
    if (ccode == STATUS_CONTINUE) {
        RemoveMboxSpaces(mbox);
        CheckPathLen(mbox);

        tmp=strchr(ptr, '(');
        if (tmp) {
            ptr=tmp;
            tmp=strchr(ptr, ')');
            if (tmp) {
                *tmp='\0';
                ccode = ParseStoreFlags(ptr + 1, &Flags);
                *tmp=')';
            }
            ptr=tmp;
        }

        tmp=ptr;  /* To remember where the date might be */

        ptr=strchr(ptr, '{'); /* fix me } */
        if (!ptr) {
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "APPEND", STATUS_INVALID_ARGUMENT));
        }

        /* Parse any date... */
        ptr2=strchr(tmp, ' ');
        if (!ptr2) {
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "APPEND", STATUS_INVALID_ARGUMENT));
        }

        /* Format is: day-month-year HH:mm:ss UTFoffset */
        while (isspace(*ptr2)) {
            ptr2++;
        }

        /* Did we get a date-time? */
        if (ptr2[0]=='"') {
            unsigned long day, month, year, hour, min, sec, offset;
            unsigned char months[20];
            unsigned long i=0;

            ptr2++;
            while (isspace(*ptr2)) {
                ptr2++;
            }
            tmp=ptr2;
            while (tmp!=ptr) {
                if (i<2 && tmp[0]=='-') {
                    tmp[0]=' ';
                    i++;;
                } else if (tmp[0]==':') {
                    tmp[0]=' ';
                }
                tmp++;
            }
 
            sscanf(ptr2, "%d %s %d %d %d %d %d", (int*)&day, months, (int *)&year, (int *)&hour, (int *)&min, (int *)&sec, (int *)&offset);

            month=1;
            for (i=0; i<12; i++) {
                if (XplStrCaseCmp(months, Imap.command.months[i])==0) {
                    month=i+1;
                    break;
                }
            }
            Date=MsgGetUTC(day, month, year, hour, min, sec);
            Date-=offset*36;
        }

        Size = atol(ptr + 1);

        if (NMAPSendCommandF(client->nmap, "STOR %s D%lu %lu %lu\r\n", mbox, Date, Flags, Size) != -1) {
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 2002) {
                if ((ConnWrite(client->conn, "+ Ready for more data\r\n", sizeof("+ Ready for more data\r\n") - 1) != -1) && (ConnFlush(client->conn) != -1)) {
                    if (ConnReadToConn(client->conn,client->nmap, Size) == Size) {
                        ccode = ReadCommandLine(client);
                        if (ccode == STATUS_CONTINUE) {
                            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                            if (ccode == 1000) {
                                /* Check Flags */
                                if (XplStrCaseCmp(client->MBox, mbox)==0) {
                                    client->NMAPAsyncEvent |= EVENT_MBOX_NEW_MESSAGES;
                                }

                                if (client->state == STATE_SELECTED) {
                                    if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                                        ;    
                                    } else {
                                        MemFree(mbox);
                                        return(SendError(client->conn, client->Ident, "APPEND", ccode));
                                    }
                                }

                                MemFree(mbox);
                                return(SendOk(client, "APPEND"));
                            }
                            MemFree(mbox);
                            return(SendError(client->conn, client->Ident, "APPEND", CheckForNMAPCommError(ccode)));
                        }
                        MemFree(mbox);
                        return(SendError(client->conn, client->Ident, "APPEND", ccode));
                    }
                }
                MemFree(mbox);
                return(STATUS_ABORT);
            }
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "APPEND", CheckForNMAPCommError(ccode)));
        }
        MemFree(mbox);
        return(SendError(client->conn, client->Ident, "APPEND", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "APPEND", ccode));
}

/********** IMAP client commands - selected state **********/

int
ImapCommandCheck(void *param)
{
    long ccode;

    ImapClient *client = (ImapClient *)param;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            return(SendOk(client, "CHECK"));
        }
        return(SendError(client->conn, client->Ident, "CHECK", ccode));
    }
    return(SendError(client->conn, client->Ident, "CHECK", STATUS_INVALID_STATE));
}

int
ImapCommandClose(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;

    if (client->state == STATE_SELECTED) {
        if (!client->ReadOnly) {
            if (NMAPSendCommand(client->nmap, "PURG\r\n", 6) == 6) {
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            } else {
                return(SendError(client->conn, client->Ident, "CLOSE", STATUS_NMAP_COMM_ERROR));
            }

            if (ccode == 1000) {
                ;
            } else {
                return(SendError(client->conn, client->Ident, "CLOSE", CheckForNMAPCommError(ccode)));
            }                   
        }

        if (NMAPSendCommand(client->nmap, "RSET MBOX\r\n", 11) == 11) {
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                /* Any last flags should get picked up in the NmapAsyncCB() */
                SendUnseenFlags(client);

                /* Any events are for the old mailbox, so forget about them. */
                client->NMAPAsyncEvent = 0;

                client->state = STATE_AUTH;
                client->MBox[0] = '\0';
 
                return(SendOk(client, "CLOSE"));
            }
            return(SendError(client->conn, client->Ident, "CLOSE", CheckForNMAPCommError(ccode)));
        } 
        return(SendError(client->conn, client->Ident, "CLOSE", STATUS_NMAP_COMM_ERROR));
    }
    return(SendError(client->conn, client->Ident, "CLOSE", STATUS_INVALID_STATE));
}

int
ImapCommandExpunge(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char Reply[1024];
    int SoFar=0;
    unsigned long imapId;


    if (client->state == STATE_SELECTED) {
        if (!(client->ReadOnly)) {
            if (NMAPSendCommand(client->nmap, "PRGV\r\n", 6) == 6) {
                ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
                while (ccode == 2002) {
                    ccode = NMAPtoNUM(client, atol(Reply), &imapId);
                    if (ccode == STATUS_CONTINUE) {
                        if (ConnWriteF(client->conn, "* %lu EXPUNGE\r\n", imapId - SoFar) != -1) {
                            SoFar++;
                            ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
                            continue;
                        }
                        return(STATUS_ABORT);
                    }

                    return(SendError(client->conn, client->Ident, "EXPUNGE", STATUS_IMAPID_NOT_FOUND));
                }

                if (ccode == 1000) {
                    ccode = LoadUIDList(client);
                    if (ccode == STATUS_CONTINUE) {
                        return(SendOk(client, "EXPUNGE"));
                    }
                    return(SendError(client->conn, client->Ident, "EXPUNGE", ccode));
                }
                return(SendError(client->conn, client->Ident, "EXPUNGE", CheckForNMAPCommError(ccode)));
            }
            return(SendError(client->conn, client->Ident, "EXPUNGE", STATUS_NMAP_COMM_ERROR));    
        }
        return(SendError(client->conn, client->Ident, "EXPUNGE", STATUS_READ_ONLY_FOLDER));
    }
    return(SendError(client->conn, client->Ident, "EXPUNGE", STATUS_INVALID_STATE));
}

__inline static long
CopyMessageRangeToTarget(ImapClient *client, unsigned long rangeStart, unsigned long rangeEnd, char *target)
{
    long ccode;
    unsigned long messageId;
    
    messageId = rangeStart;

    do {
        /* Do the dirty work */
        if (NMAPSendCommandF(client->nmap, "COPY %lu %s\r\n", client->IDList[messageId], target) != -1) {
            ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
            if (ccode == 1000) {
                if (messageId < rangeEnd) {
                    messageId++;
                    continue;
                }

                break;
            }

            if (ccode == 4224) {
                return(STATUS_TRY_CREATE);
            }

            return(CheckForNMAPCommError(ccode));
        }
        return(STATUS_NMAP_COMM_ERROR);
    } while (TRUE);

    return(STATUS_CONTINUE);
}

__inline static long
HandleCopy(ImapClient *client, BOOL ByUID)
{
    char *nextRange;
    unsigned long rangeStart;
    unsigned long rangeEnd;
    
    unsigned char *messageSet;
    unsigned char *ptr;
    
    unsigned char *Target;
    unsigned char *Command;
    long ccode;
 
    if (ByUID) {
        Command="UID COPY";
    } else {
        Command="COPY";
    }

    ptr = client->Command + 5;

    /* Grab the range */
    ccode = GrabArgument(client, &ptr, &messageSet);
    if (ccode == STATUS_CONTINUE) {
        /* Grab the target mailbox */
        ccode = GrabArgument(client, &ptr, &Target);
        if (ccode == STATUS_CONTINUE) {
            RemoveMboxSpaces(Target);
            CheckPathLen(Target);

            nextRange = messageSet;
            do {
                ccode = GetMessageRange(client, &(nextRange), &rangeStart, &rangeEnd, ByUID);
                if (ccode == STATUS_CONTINUE) {
                    ccode = CopyMessageRangeToTarget(client, rangeStart, rangeEnd, Target);
                    if (ccode == STATUS_CONTINUE) {
                        continue;
                    }
                }
                if (ccode == STATUS_UID_NOT_FOUND) {
                    continue;
                }
                MemFree(messageSet);
                MemFree(Target);
                return(SendError(client->conn, client->Ident, Command, ccode));
            } while (nextRange);

            MemFree(messageSet);
            MemFree(Target);
            return(SendOk(client, Command));
        }
        MemFree(messageSet);
    }
    return(SendError(client->conn, client->Ident, Command, ccode));
}

int
ImapCommandCopy(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            return(HandleCopy(client, FALSE));    
        }
        
        return(SendError(client->conn, client->Ident, "COPY", ccode));
    }

    return(SendError(client->conn, client->Ident, "COPY", STATUS_INVALID_STATE));
}

int
ImapCommandUidCopy(void *param)
{
    long ccode;
    ImapClient *client = (ImapClient *)param;
    unsigned char *ptr;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ptr = client->Command + 4;
            memmove(client->Command, ptr, strlen(ptr) + 1);
            return(HandleCopy(client, TRUE));
        }

        return(SendError(client->conn, client->Ident, "UID COPY", ccode));
    }

    return(SendError(client->conn, client->Ident, "UID COPY", STATUS_INVALID_STATE));
}

/********** IMAP ACL commands rfc2086 **********/

int
ImapCommandSetAcl(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Answer[1024];
    unsigned long permissions;
    unsigned long newPermissions;
    unsigned long action;
    unsigned char *ptr2;
    unsigned char *mbox;
    unsigned char *mailbox;
    unsigned char *identifier;
    BOOL    ready;

    if (Imap.command.capability.acl.enabled) {
        if (client->state == STATE_SELECTED) {
            if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                ;    
            } else {
                return(SendError(client->conn, client->Ident, "SETACL", ccode));
            }
        } else if (client->state < STATE_AUTH) { 
            return(SendError(client->conn, client->Ident, "SETACL", STATUS_INVALID_STATE));
        }

        ptr = client->Command + 6;
        if (isspace(*ptr)) {
            do { ptr++; } while (isspace(*ptr));
        } else {
            return(SendError(client->conn, client->Ident, "SETACL", STATUS_INVALID_ARGUMENT));
        }

        /* Grab the mailbox name. */
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            /* Grab the authentication identifier */
            ccode = GrabArgument(client, &ptr, &identifier);
            if (ccode == STATUS_CONTINUE) {
                /* Advance as needed. */
                if (isspace(*ptr)) {
                    do { ptr++; } while (isspace(*ptr));
                }

                switch (*ptr) {
                case '\0': { action = 0;         break; }
                case '-': {  action = -1; ptr++;     break; }
                case '+': {  action = 1;  ptr++;     break; }
                default: {  action = 0;         break; }
                }

                /* Advance as needed. */
                if (isspace(*ptr)) {
                    do { ptr++; } while (isspace(*ptr));
                }

                ready = TRUE;
                newPermissions = 0;
                while ((ready == TRUE) && (*ptr != '\0')) {
                    switch(toupper(*ptr)) {
                    case 'A': { /* Administer (perform SETACL) */
                        newPermissions |= NMAP_SHARE_ADMINISTER; break; }

                    case 'C': { /* Create (CREATE new sub-mailboxes in any implementation-defined hierarchy. */
                        newPermissions |= NMAP_SHARE_CREATE;  break; }

                    case 'D': { /* Delete (STORE DELETED flag, perform EXPUNGE). */
                        newPermissions |= NMAP_SHARE_DELETE;  break; }

                    case 'I': { /* Insert (perform APPEND, COPY into mailbox). */
                        newPermissions |= NMAP_SHARE_INSERT;  break; }

                    case 'L': { /* Lookup (mailbox is visible to LIST/LSUB commands). */
                        newPermissions |= NMAP_SHARE_VISIBLE;  break; }

                    case 'P': { /* Post (send mail to submission address for mailbox, not enforced by IMAP4 itself. */
                        newPermissions |= NMAP_SHARE_POST;   break; }

                    case 'R': { /* Read (SELECT the mailbox, perform CHECK, FETCH, PARTIAL, SEARCH, COPY from mailbox. */
                        newPermissions |= NMAP_SHARE_READ;   break; }

                    case 'S': { /* Keep seen/unseen information across sessions (STORE SEEN flag) */
                        newPermissions |= NMAP_SHARE_SEEN;   break; }

                    case 'W': { /* Write (STORE flags other than SEEN and DELETED). */
                        newPermissions |= NMAP_SHARE_WRITE;   break; }

                    default: {
                        MemFree(mbox);
                        MemFree(identifier);

                        ready = FALSE;
                        break;
                    }
                    }

                    do { ptr++; } while(isspace(*ptr));
                }

                if (ready == TRUE) {
                    mailbox = mbox;
                    if ((*mailbox == '/') || (*mailbox == '\\')) {
                        mailbox++;
                    }

                    RemoveMboxSpaces(mailbox);
                    CheckPathLen(mailbox);
                } else {
                    /* Invalid permissions! */
                    return(SendError(client->conn, client->Ident, "SETACL", STATUS_PERMISSION_DENIED));
                }

                if (action != 0) {
                    permissions = 0;

                    /* Load the NMAP permissions for this resource. */
                    if (NMAPSendCommand(client->nmap, "SHARE MBOX\r\n", 12) == 12) {
                        ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);

                        if (ccode == 2002) {
                            ready = FALSE;
                            while ((ccode = NMAPReadResponse(client->nmap, Answer, sizeof(Answer), TRUE)) == 2002) {
                                if (ready == FALSE) {
                                    ptr = strchr(Answer, ' ');
                                    if (ptr != NULL) {
                                        *(ptr++) = '\0';

                                        if (XplStrCaseCmp(mailbox, Answer) != 0) {
                                            continue;
                                        }
                                    } else {
                                        continue;
                                    }

                                    ptr2 = strchr(ptr, ' ');
                                    if (ptr2 != NULL) {
                                        *(ptr2++) = '\0';           

                                        if (XplStrCaseCmp(identifier, ptr) != 0) {
                                            continue;
                                        }
                                    } else {
                                        continue;
                                    }

                                    permissions = atol(ptr2);
                                    ready = TRUE;

                                    /* Keep looping to absorb the remaining NMAP Server responses. */
                                }
                            }

                            if (ccode == 1000) {
                                ;
                            } else {
                                MemFree(mbox);
                                MemFree(identifier);
                                return(SendError(client->conn, client->Ident, "SETACL", CheckForNMAPCommError(ccode)));
                            }
                        } else {
                            MemFree(mbox);
                            MemFree(identifier);
                            return(SendError(client->conn, client->Ident, "SETACL", CheckForNMAPCommError(ccode)));
                        }

                        if (action == 1) {
                            ccode = NMAPSendCommandF(client->nmap, "SHARE MBOX %s %s %lu%s\r\n", mailbox, identifier, permissions | newPermissions, (ready == TRUE)? " NOMESSAGE": "");
                        } else {
                            ccode = NMAPSendCommandF(client->nmap, "SHARE MBOX %s %s %lu%s\r\n", mailbox, identifier, permissions & ~newPermissions, (ready == TRUE)? " NOMESSAGE": "");
                        }

                        if (ccode != -1) {
                            ;
                        } else {
                            MemFree(mbox);
                            MemFree(identifier);
                            return(SendError(client->conn, client->Ident, "SETACL", STATUS_NMAP_COMM_ERROR));
                        }
                    } else {
                        MemFree(mbox);
                        MemFree(identifier);
                        return(SendError(client->conn, client->Ident, "SETACL", STATUS_NMAP_COMM_ERROR));
                    }
                } else {
                    /* Override existing permissions. */
                    if (NMAPSendCommandF(client->nmap, "SHARE MBOX %s %s %lu NOMESSAGE\r\n", mailbox, identifier, newPermissions) != -1) {
                        ;
                    } else {
                        MemFree(mbox);
                        MemFree(identifier);
                        return(SendError(client->conn, client->Ident, "SETACL", STATUS_NMAP_COMM_ERROR));
                    }
                }

                MemFree(mbox);
                MemFree(identifier);

                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode == 1000) {
                    return(SendOk(client, "SETACL"));
                }
                return(SendError(client->conn, client->Ident, "SETACL", CheckForNMAPCommError(ccode)));
            }
        }
        return(SendError(client->conn, client->Ident, "SETACL", ccode));
    }
    return(SendError(client->conn, client->Ident, "SETACL", STATUS_FEATURE_DISABLED));
}

int
ImapCommandDeleteAcl(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char *mbox;
    unsigned char *mailbox;
    unsigned char *identifier;

    if (Imap.command.capability.acl.enabled) {
        if (client->state == STATE_SELECTED) {
            if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                ;    
            } else {
                return(SendError(client->conn, client->Ident, "DELETEACL", ccode));
            }
        } else if (client->state < STATE_AUTH) {
            return(SendError(client->conn, client->Ident, "DELETEACL", STATUS_INVALID_STATE));
        }

        ptr = client->Command + 9;
        if (isspace(*ptr)) {
            do { ptr++; } while (isspace(*ptr));
        } else {
            return(SendError(client->conn, client->Ident, "DELETEACL", STATUS_INVALID_ARGUMENT));
        }

        /* Grab the mailbox name. */
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            /* Grab the authentication identifier */
            ccode = GrabArgument(client, &ptr, &identifier);
            if (ccode == STATUS_CONTINUE) {
                mailbox = mbox;
                if ((*mailbox == '/') || (*mailbox == '\\')) {
                    mailbox++;
                }

                RemoveMboxSpaces(mailbox);
                CheckPathLen(mailbox);

                if (NMAPSendCommandF(client->nmap, "SHARE MBOX %s %s 0 NOMESSAGE\r\n", mailbox, identifier) != -1) {
                    MemFree(mbox);
                    MemFree(identifier);
    
                    ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                    if (ccode == 1000) {
                        return(SendOk(client, "DELETEACL"));
                    }
                    return(SendError(client->conn, client->Ident, "DELETEACL", CheckForNMAPCommError(ccode)));
                }
                MemFree(mbox);
                MemFree(identifier);
                return(SendError(client->conn, client->Ident, "DELETEACL", STATUS_NMAP_COMM_ERROR));
            }
        }
        return(SendError(client->conn, client->Ident, "DELETEACL", ccode));
    }
    return(SendError(client->conn, client->Ident, "DELETEACL", STATUS_FEATURE_DISABLED));
}

int
ImapCommandGetAcl(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Answer[1024];
    unsigned char Reply[1024];
    unsigned long permissions;
    unsigned char *ptr2;
    unsigned char *mbox;
    unsigned char *mailbox;

    if (Imap.command.capability.acl.enabled) {
        if (client->state == STATE_SELECTED) {
            if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                ;    
            } else {
                return(SendError(client->conn, client->Ident, "GETACL", ccode));
            }
        } else if (client->state < STATE_AUTH) {
            return(SendError(client->conn, client->Ident, "GETACL", STATUS_INVALID_STATE));
        }

        ptr = client->Command + 6;
        if (isspace(*ptr)) {
            do { ptr++; } while (isspace(*ptr));
        } else {
            return(SendError(client->conn, client->Ident, "GETACL", STATUS_INVALID_ARGUMENT));
        }

        /* Grab the mailbox name. */
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            mailbox = mbox;
            if ((*mailbox == '/') || (*mailbox == '\\')) {
                mailbox++;
            }

            ptr2 = Reply + snprintf(Reply, sizeof(Reply), "* ACL \"%s\"", mailbox);

            RemoveMboxSpaces(mailbox);
            CheckPathLen(mailbox);

            if (NMAPSendCommandF(client->nmap, "RIGHTS %s\r\n", mailbox) != -1) {
                MemFree(mbox);
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode == 2002) {
                    while ((ccode = NMAPReadResponse(client->nmap, Answer, sizeof(Answer), TRUE)) == 2002) {
                        ptr = strchr(Answer, ' ');
                        if (ptr != NULL) {
                            *(ptr++) = '\0';

                            permissions = atol(ptr);

                            /* Add the grantee name to the response. */
                            ptr2 += snprintf(ptr2, sizeof(Reply) - (ptr2 - Reply), " \"%s\" ", Answer);
                            if (permissions) {
                                if (permissions & NMAP_SHARE_VISIBLE) {
                                    *(ptr2++) = 'l';
                                }

                                if (permissions & NMAP_SHARE_READ) {
                                    *(ptr2++) = 'r';
                                }

                                if (permissions & NMAP_SHARE_SEEN) {
                                    *(ptr2++) = 's';
                                }

                                if (permissions & NMAP_SHARE_WRITE) {
                                    *(ptr2++) = 'w';
                                }

                                if (permissions & NMAP_SHARE_INSERT) {
                                    *(ptr2++) = 'i';
                                }

                                if (permissions & NMAP_SHARE_POST) {
                                    *(ptr2++) = 'p';
                                }

                                if (permissions & NMAP_SHARE_CREATE) {
                                    *(ptr2++) = 'c';
                                }

                                if (permissions & NMAP_SHARE_DELETE) {
                                    *(ptr2++) = 'd';
                                }

                                if (permissions & NMAP_SHARE_ADMINISTER) {
                                    *(ptr2++) = 'a';
                                }

                                continue;
                            } else {
                                *(ptr2++) = '"';
                                *(ptr2++) = '"';
                            }
                        }

                        continue;
                    }

                    if (ccode == 1000) {
                        *(ptr2++) = '\r';
                        *(ptr2++) = '\n';
                        *(ptr2) = '\0';

                        if (ConnWrite(client->conn, Reply, ptr2 - Reply) != -1) {
                            return(SendOk(client, "GETACL"));
                        }
                        return(STATUS_ABORT);
                    }
                }
                return(SendError(client->conn, client->Ident, "GETACL", CheckForNMAPCommError(ccode)));
            }
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "GETACL", STATUS_NMAP_COMM_ERROR));
        }
        return(SendError(client->conn, client->Ident, "GETACL", ccode));
    }
    return(SendError(client->conn, client->Ident, "GETACL", STATUS_FEATURE_DISABLED));
}

int
ImapCommandListRights(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *mbox;
    unsigned char *identifier;
    unsigned char *ptr;

    if (Imap.command.capability.acl.enabled) {
        if (client->state == STATE_SELECTED) {
            if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                ;    
            } else {
                return(SendError(client->conn, client->Ident, "LISTRIGHTS", ccode));
            }
        }

        if (client->state >= STATE_AUTH) {
            ptr = client->Command + 10;
            if (isspace(*ptr)) {
                do { ptr++; } while (isspace(*ptr));
            } else {
                return(SendError(client->conn, client->Ident, "LISTRIGHTS", STATUS_INVALID_ARGUMENT));
            }
        } else {
            return(SendError(client->conn, client->Ident, "LISTRIGHTS", STATUS_INVALID_STATE));
        }

        /* Grab the mailbox name. */
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            identifier = NULL;

            /* Grab the authentication identifier */
            ccode = GrabArgument(client, &ptr, &identifier);
            if (ccode == STATUS_CONTINUE) {
                MemFree(mbox);
                MemFree(identifier);
                if (ConnWriteF(client->conn, "* LISTRIGHTS \"%s\" %s \"\" l r s w i p c d a\r\n", mbox, identifier) != -1) {
                    return(SendOk(client, "LISTRIGHTS"));
                }
                return(STATUS_ABORT);
            }
            MemFree(mbox);
        }
        return(SendError(client->conn, client->Ident, "LISTRIGHTS", ccode));
    }
    return(SendError(client->conn, client->Ident, "LISTRIGHTS", STATUS_FEATURE_DISABLED));
}

int
ImapCommandMyRights(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Answer[1024];
    unsigned char Reply[1024];
    unsigned long permissions;
    unsigned char *ptr2;
    unsigned char *mbox;
    unsigned char *mailbox;
    BOOL found;

    if (Imap.command.capability.acl.enabled) {
        if (client->state == STATE_SELECTED) {
            if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
                ;    
            } else {
                return(SendError(client->conn, client->Ident, "MYRIGHTS", ccode));
            }
        } else if (client->state < STATE_AUTH) {
            return(SendError(client->conn, client->Ident, "MYRIGHTS", STATUS_INVALID_STATE));
        }

        ptr = client->Command + 8;
        if (isspace(*ptr)) {
            do { ptr++; } while (isspace(*ptr));
        } else {
            return(SendError(client->conn, client->Ident, "MYRIGHTS", STATUS_INVALID_ARGUMENT));
        }

        /* Grab the mailbox name. */
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            mailbox = mbox;
            if ((*mailbox == '/') || (*mailbox == '\\')) {
                mailbox++;
            }

            ptr2 = Reply + snprintf(Reply, sizeof(Answer), "* MYRIGHTS \"%s\" \"\"", mailbox);

            RemoveMboxSpaces(mailbox);
            CheckPathLen(mailbox);

            if (NMAPSendCommandF(client->nmap, "RIGHTS %s\r\n", mailbox) != -1) {
                MemFree(mbox);
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode == 2002) {
                    found = FALSE;
                    while ((ccode = NMAPReadResponse(client->nmap, Answer, sizeof(Answer), TRUE)) == 2002) {
                        ptr = strchr(Answer, ' ');
                        if (ptr != NULL) {
                            *(ptr++) = '\0';

                            /* We are only interested in this user's permissions.
                               Keep looping to absorb all NMAP responses. */
                            if (XplStrCaseCmp(client->User, Answer) == 0) {
                                permissions = atol(ptr);
                                if (permissions) {
                                    ptr2 -= 2; /* Back over the default empty permissions. */

                                    if (permissions & NMAP_SHARE_VISIBLE) {
                                        *(ptr2++) = 'l';
                                    }

                                    if (permissions & NMAP_SHARE_READ) {
                                        *(ptr2++) = 'r';
                                    }

                                    if (permissions & NMAP_SHARE_SEEN) {
                                        *(ptr2++) = 's';
                                    }

                                    if (permissions & NMAP_SHARE_WRITE) {
                                        *(ptr2++) = 'w';
                                    }

                                    if (permissions & NMAP_SHARE_INSERT) {
                                        *(ptr2++) = 'i';
                                    }

                                    if (permissions & NMAP_SHARE_POST) {
                                        *(ptr2++) = 'p';
                                    }

                                    if (permissions & NMAP_SHARE_CREATE) {
                                        *(ptr2++) = 'c';
                                    }

                                    if (permissions & NMAP_SHARE_DELETE) {
                                        *(ptr2++) = 'd';
                                    }

                                    if (permissions & NMAP_SHARE_ADMINISTER) {
                                        *(ptr2++) = 'a';
                                    }

                                    continue;
                                }
                            }

                            continue;
                        }
                    }

                    if (ccode == 1000) {
                        *(ptr2++) = '\r';
                        *(ptr2++) = '\n';
                        *(ptr2) = '\0';
                        if (ConnWrite(client->conn, Reply, ptr2 - Reply) != -1) {
                            return(SendOk(client, "MYRIGHTS"));
                        }
                        return(STATUS_ABORT);
                    }
                }
                return(SendError(client->conn, client->Ident, "MYRIGHTS", CheckForNMAPCommError(ccode)));
            }
            MemFree(mbox);
            return(SendError(client->conn, client->Ident, "MYRIGHTS", STATUS_NMAP_COMM_ERROR));
        }
        return(SendError(client->conn, client->Ident, "MYRIGHTS", ccode));
    }
    return(SendError(client->conn, client->Ident, "MYRIGHTS", STATUS_FEATURE_DISABLED));
}

/********** IMAP Quota commands rfc2087 **********/

int
ImapCommandSetQuota(void *param)
{
    long ccode;

    ImapClient *client = (ImapClient *)param;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "SETQUOTA", ccode));
        }
    }

    return(SendError(client->conn, client->Ident, "SETQUOTA", STATUS_PERMISSION_DENIED));
}

int
ImapCommandGetQuota(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];
    unsigned long quota = 0;
    unsigned long used = 0;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "GETQUOTA", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_INVALID_STATE));
    }

    if (NMAPSendCommandF(client->nmap, "SPACE\r\n") != -1) {
        ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
        if(ccode != 1000) {
            ccode = CheckForNMAPCommError(ccode);
            if (ccode != STATUS_NMAP_COMM_ERROR) {
                return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_COULD_NOT_DETERMINE_QUOTA));
            }
            return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_NMAP_COMM_ERROR));
        }

        ptr = strchr(Reply, ' ');
        if(!ptr) {
            return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_COULD_NOT_DETERMINE_QUOTA));
        }
 
        quota = atol(ptr);
        used = atol(Reply);

        if(client->Command[9] == '\"' && client->Command[10] == '\"') {
            if (ConnWriteF(client->conn, "* QUOTA \"\" (STORAGE %lu %lu)\r\n%s OK GETQUOTA completed\r\n", used, quota, client->Ident) != -1) {
                return(STATUS_CONTINUE);
            }
            return(STATUS_ABORT);
        }
        return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_NO_SUCH_QUOTA_ROOT));
    }
    return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_NMAP_COMM_ERROR));
}

int
ImapCommandGetQuotaRoot(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];
    unsigned long quota = 0;
    unsigned long used = 0;
    unsigned char *mbox;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "QUOTAROOT", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "GETQUOTAROOT", STATUS_INVALID_STATE));
    }

    if (NMAPSendCommandF(client->nmap, "SPACE\r\n") != -1) {
        ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
        if(ccode != 1000) {
            ccode = CheckForNMAPCommError(ccode);
            if (ccode != STATUS_NMAP_COMM_ERROR) {
                return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_COULD_NOT_DETERMINE_QUOTA));
            }
            return(SendError(client->conn, client->Ident, "GETQUOTA", STATUS_NMAP_COMM_ERROR));
        }
        ptr = strchr(Reply, ' ');
        if(!ptr) {
            return(SendError(client->conn, client->Ident, "GETQUOTAROOT", STATUS_COULD_NOT_DETERMINE_QUOTA));
        }
        quota = atol(ptr);
        used = atol(Reply);


        ptr = client->Command + 13;
        if(*(ptr - 1) != ' ') {
            return(SendError(client->conn, client->Ident, "GETQUOTAROOT", STATUS_INVALID_ARGUMENT));
        }
        ccode = GrabArgument(client, &ptr, &mbox);
        if (ccode == STATUS_CONTINUE) {
            ccode = ConnWriteF(client->conn, "* QUOTAROOT %s \"\"\r\n* QUOTA \"\" (STORAGE %lu %lu)\r\n%s OK GETQUOTAROOT completed\r\n", mbox, used, quota, client->Ident);
            MemFree(mbox);
            if (ccode != -1) {
                return(STATUS_CONTINUE);
            }
            return(STATUS_ABORT);
        }
        return(SendError(client->conn, client->Ident, "QUOTAROOT", ccode));
    }
    return(SendError(client->conn, client->Ident, "QUOTAROOT", STATUS_NMAP_COMM_ERROR));
}

/********** IMAP Namespace command rfc2342 **********/

int
ImapCommandNameSpace(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char Reply[1024];

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "NAMESPACE", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "NAMESPACE", STATUS_INVALID_STATE));
    }

    if (NMAPSendCommand(client->nmap, "NAMESPACE\r\n", 11) == 11) {
        ccode = NMAPReadResponse(client->nmap, Reply, sizeof(Reply), TRUE);
        if (ccode == 1000) {
            unsigned char delim;
            unsigned char *user;
            unsigned char *userShare;
            unsigned char *publicShare;

            /* "/" "" "Shares/" "Public Shares/" */
            /* (("" "/")) (("Shares/" "/")) (("Public Shares/" "/")) */
            ptr = Reply;
            while (isspace(*ptr)) {
                ptr++;
            }

            delim = *(++ptr);
            ptr += 2;

            while (*ptr != '"') {
                ptr++;
            }

            user = ++ptr;
            while (*ptr != '"') {
                ptr++;
            }
            *ptr = '\0';
            ptr += 2;
            while (*ptr != '"') {
                ptr++;
            }

            userShare = ++ptr;
            while (*ptr != '"') {
                ptr++;
            }
            *ptr = '\0';
            ptr += 2;
            while (*ptr != '"') {
                ptr++;
            }

            publicShare = ++ptr;
            while (*ptr != '"') {
                ptr++;
            }
            *ptr = '\0';

            if (ConnWriteF(client->conn, "* NAMESPACE ((\"%s\" \"%c\")) ((\"%s\" \"%c\")) ((\"%s\" \"%c\"))\r\n", user, delim, userShare, delim, publicShare, delim) != -1) {
                return(SendOk(client, "NAMESPACE"));
            }
            return(STATUS_ABORT);
        }
        return(SendError(client->conn, client->Ident, "NAMESPACE", CheckForNMAPCommError(ccode)));
    }
    return(SendError(client->conn, client->Ident, "NAMESPACE", STATUS_NMAP_COMM_ERROR));
}

/********** non-standard commands **********/

int
ImapCommandProxyAuth(void *param)
{
    ImapClient *client = (ImapClient *)param;
    long ccode;
    unsigned char *ptr;
    unsigned char  *user = NULL;
    MDBValueStruct  *User;

    if (client->state == STATE_SELECTED) {
        if (((ccode = PurgeMessageNotify(client)) == STATUS_CONTINUE) && ((ccode = NewMessageNotify(client)) == STATUS_CONTINUE)) {
            ;    
        } else {
            return(SendError(client->conn, client->Ident, "PROXYAUTH", ccode));
        }
    } else if (client->state < STATE_AUTH) {
        return(SendError(client->conn, client->Ident, "PROXYAUTH", STATUS_INVALID_STATE));
    }

    /* One of the consequences of switching users the way we do is that proxyauth
     * will only succeed the first time. After that, the user has changed and can
     * no longer proxyauth. This is the way Netscape does it, so supposedly its
     * correct, although inefficient for the client. */
    if(client->state < STATE_AUTH || XplStrCaseCmp(client->User, Imap.server.postmaster)) {
        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_UNHANDLED, LOGGER_EVENT_UNHANDLED_REQUEST, LOG_INFO, 0, client->User, client->Command, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
        return(SendError(client->conn, client->Ident, "", STATUS_UNKNOWN_COMMAND));
    }
    ptr = client->Command + 10;
    while(isspace(*ptr)) {
        ++ptr;
    }

    if(!*ptr) {
        return(SendError(client->conn, client->Ident, "", STATUS_INVALID_ARGUMENT));
    }

    ccode = GrabArgument(client, &ptr, &user);
    if (ccode == STATUS_CONTINUE) {
        /* Fixme - this doesn't connect the user to *his/her* nmap server */

        User = MDBCreateValueStruct(Imap.directory.handle, NULL);
        if (MsgFindObject(user, NULL, NULL, NULL, User)) { 
            if(XplStrCaseCmp(client->User, User->Value[0]) != 0) {
                if (strlen(User->Value[0]) < sizeof(client->User)) {
                    strcpy(client->User, User->Value[0]);
                }
            }
            MemFree(user);
            MDBDestroyValueStruct(User);

            client->state = STATE_AUTH;
            if (NMAPSendCommandF(client->nmap, "USER %s\r\n", client->User) != -1) {
                ccode = NMAPReadResponse(client->nmap, NULL, 0, 0);
                if (ccode == 1000) {
                    return(SendOk(client, "PROXYAUTH"));
                }
                return(SendError(client->conn, client->Ident, "PROXYAUTH", CheckForNMAPCommError(ccode)));
            }
            return(SendError(client->conn, client->Ident, "PROXYAUTH", STATUS_NMAP_COMM_ERROR));
        }

        MDBDestroyValueStruct(User);
        MemFree(user);
        return(SendError(client->conn, client->Ident, "PROXYAUTH", STATUS_USERNAME_NOT_FOUND));
    }

    return(SendError(client->conn, client->Ident, "PROXYAUTH", ccode));
}

static BOOL
HandleConnection(void *param)
{
    long ccode;
    long commandId;
    ImapClient  *client=(ImapClient *)param;
    unsigned char *ptr;
    ProtocolCommand *localCommand;

    if (ConnNegotiate(client->conn, Imap.server.ssl.context)) {
        if (client->conn->ssl.enable == FALSE) {
            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_CONNECTION, LOG_INFO, 0, NULL, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
        } else {
            LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_AUTH, LOGGER_EVENT_SSL_CONNECTION, LOG_INFO, 0, NULL, NULL, XplHostToLittle(client->conn->socketAddress.sin_addr.s_addr), 0, NULL, 0);
        }

        if ((ConnWriteF(client->conn, "* OK %s %s server ready <%ud.%lu@%s>\r\n",Imap.server.hostname, PRODUCT_NAME, XplGetThreadID(),time(NULL),Imap.server.hostname) != -1) && (ConnFlush(client->conn) != -1)) {
            while (TRUE) {
                ccode = ReadCommandLine(client);
                if (ccode == STATUS_CONTINUE) {
                    /* Grab ident from command */
                    ptr = strchr(client->Command,' ');
                    if (ptr) {
                        if ((unsigned long)(ptr - client->Command) < sizeof(client->Ident)) {
                            ptr[0]='\0';
                            memcpy(client->Ident, client->Command, ptr - client->Command + 1);
                        } else {
                            memcpy(client->Ident, client->Command, sizeof(client->Ident));
                            client->Ident[sizeof(client->Ident)-1]='\0';
                        }
                        memmove(client->Command, ptr + 1, strlen(ptr + 1) + 1);
                        commandId = HulaKeywordFind(Imap.command.index, client->Command);
                        if (commandId != -1) {
                            localCommand = &ImapProtocolCommands[commandId];
                            ccode = localCommand->handler(client);
                        } else {
                            ccode = ImapCommandUnknown(client);
                        }

                        if (ccode != STATUS_ABORT) {
                            /* Post-Command Processing */
                            if (client->state == STATE_SELECTED) {
                                SendUnseenFlags(client);
                            }
                            ConnFlush(client->conn);
                            continue;
                        }

                        break;
                    }

                    if ((ConnWrite(client->conn, "* BAD Missing command\r\n", 23) != -1) && (ConnFlush(client->conn) != -1)) {
                        continue;
                    }
                    break;
                }

                if ((ccode != STATUS_ABORT) && (ConnWrite(client->conn, "* BAD command line too long\r\n", 29) != -1) && (ConnFlush(client->conn) != -1)) {
                    continue;
                }
                break;
            }
        }
    }

    SessionCleanup(client);
    return(FALSE);
}



#if defined(NETWARE) || defined(LIBC) || defined(WIN32)
static int 
_NonAppCheckUnload(void)
{
    int   s;
    static BOOL checked = FALSE;
    XplThreadID id;

    if (!checked) {
        checked = Imap.exiting = TRUE;

        XplWaitOnLocalSemaphore(Imap.server.shutdownSemaphore);

        id = XplSetThreadGroupID(Imap.server.threadGroupId);
        if (Imap.server.conn->socket != -1) {
            s = Imap.server.conn->socket;
            Imap.server.conn->socket = -1;

            IPclose(s);
        }
        XplSetThreadGroupID(id);

        XplWaitOnLocalSemaphore(Imap.server.mainSemaphore);
    }

    return(0);
}
#endif

static void 
IMAPShutdownSigHandler(int sigtype)
{
    static BOOL signaled = FALSE;

    if (!signaled && (sigtype == SIGTERM || sigtype == SIGINT)) {
        signaled = Imap.exiting = TRUE;

        if (Imap.server.conn) {
            ConnClose(Imap.server.conn, 1);
            Imap.server.conn = NULL;
        }

        XplSignalLocalSemaphore(Imap.server.shutdownSemaphore);

        /* Allow the main thread to close the final semaphores and exit. */
        XplDelay(1000);
    }

    return;
}

static int
ServerSocketInit(void)
{
    Imap.server.conn = ConnAlloc(FALSE);
    if (!Imap.server.conn) {
        XplConsolePrintf("hulaimap: Could not allocate connection.\n");
        return -1;
    }

    Imap.server.conn->socketAddress.sin_family = AF_INET;
    Imap.server.conn->socketAddress.sin_port = htons(Imap.server.port);
    Imap.server.conn->socketAddress.sin_addr.s_addr = MsgGetAgentBindIPAddress();

    /* Get root privs back for the bind.  It's ok if this fails - 
     * the user might not need to be root to bind to the port */
    XplSetEffectiveUserId(0);
    
    Imap.server.conn->socket = ConnServerSocket(Imap.server.conn, 2048);

    if (XplSetEffectiveUser(MsgGetUnprivilegedUser()) < 0) {
        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_PRIV_FAILURE, LOG_ERROR, 0, MsgGetUnprivilegedUser(), NULL, 0, 0, NULL, 0);
        XplConsolePrintf("hulaimap: Could not drop to unprivileged user '%s'\n", MsgGetUnprivilegedUser());
        return -1;
    }

    if (Imap.server.conn->socket < 0) {
        int ret = Imap.server.conn->socket;
        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_BIND_FAILURE, LOG_ERROR, 0, "IMAP ", NULL, Imap.server.port, 0, NULL, 0);
        XplConsolePrintf("hulaimap: Could not bind to port %d\n", Imap.server.port);
        ConnFree(Imap.server.conn);
        return ret;
    }

    return 0;
}

static int
ServerSocketSSLInit(void)
{
    Imap.server.ssl.conn = ConnAlloc(FALSE);
    if (Imap.server.ssl.conn == NULL) {
        XplConsolePrintf("hulaimap: Could not allocate connection.\n");
        ConnFree(Imap.server.conn);
        return -1;
    }

    Imap.server.ssl.conn->socketAddress.sin_family = AF_INET;
    Imap.server.ssl.conn->socketAddress.sin_port = htons(Imap.server.ssl.port);
    Imap.server.ssl.conn->socketAddress.sin_addr.s_addr = MsgGetAgentBindIPAddress();
        
    /* Get root privs back for the bind.  It's ok if this fails - 
     * the user might not need to be root to bind to the port */
    XplSetEffectiveUserId(0);
    
    Imap.server.ssl.conn->socket = ConnServerSocket(Imap.server.ssl.conn, 2048);

    if (XplSetEffectiveUser(MsgGetUnprivilegedUser()) < 0) {
        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_PRIV_FAILURE, LOG_ERROR, 0, MsgGetUnprivilegedUser(), NULL, 0, 0, NULL, 0);
        XplConsolePrintf("hulaimap: Could not drop to unprivileged user '%s'\n", MsgGetUnprivilegedUser());
        return -1;
    }
        
    if (Imap.server.ssl.conn->socket < 0) {
        int ret = Imap.server.ssl.conn->socket;

        LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_BIND_FAILURE, LOG_ERROR, 0, "IMAP-SSL ", NULL, Imap.server.ssl.port, 0, NULL, 0);
        XplConsolePrintf("hulaimap: Could not bind to SSL port %d\n", Imap.server.ssl.port);
        ConnFree(Imap.server.conn);
        ConnFree(Imap.server.ssl.conn);
        
        return ret;
    }

    return 0;
}

static BOOL
InitializeImapGlobals()
{
    /*      Imap.client.        */
    XplSafeWrite(Imap.client.threads.inUse, 0);
    XplSafeWrite(Imap.client.threads.idle, 0);
    XplSafeWrite(Imap.client.served, 0);
    XplSafeWrite(Imap.client.badPassword, 0);
    Imap.client.threads.max = 100000;

    /*      Imap.server.        */
    XplSafeWrite(Imap.server.active, 0);
    Imap.server.disabled = FALSE;
    Imap.server.ssl.enable = FALSE;
    Imap.server.threadGroupId = XplGetThreadGroupID();
    Imap.server.hostname[0] = '\0';
    Imap.server.hostnameLen = 0;
    strcpy(Imap.server.postmaster, "admin");

    /*      Imap.command.        */    
    Imap.command.capability.acl.enabled = TRUE;
    Imap.command.capability.len = sprintf(Imap.command.capability.message, "%s\r\n", "* CAPABILITY IMAP4 IMAP4rev1 ACL AUTH=LOGIN NAMESPACE XSENDER");
    Imap.command.capability.ssl.len = sprintf(Imap.command.capability.ssl.message, "%s\r\n", "* CAPABILITY IMAP4 IMAP4rev1 ACL AUTH=LOGIN NAMESPACE STARTTLS XSENDER LOGINDISABLED");

    Imap.command.months[0] = "Jan";
    Imap.command.months[1] = "Feb";
    Imap.command.months[2] = "Mar";
    Imap.command.months[3] = "Apr";
    Imap.command.months[4] = "May";
    Imap.command.months[5] = "Jun";
    Imap.command.months[6] = "Jul";
    Imap.command.months[7] = "Aug";
    Imap.command.months[8] = "Sep";
    Imap.command.months[9] = "Oct";
    Imap.command.months[10] = "Nov";
    Imap.command.months[11] = "Dec";

    /*      Imap.(misc).        */    
    Imap.exiting = FALSE;
    Imap.logHandle = NULL;
    Imap.directory.handle = NULL;

    /* Global allocations */
    HulaKeywordIndexCreateFromTable(Imap.command.index, ImapProtocolCommands, .name, TRUE);
    if (Imap.command.index) {
        Imap.command.returnValueIndex = CreateReturnValueIndex();
        if (Imap.command.returnValueIndex) {
            if (CommandFetchInit()) {
                if (CommandStoreInit()) {
                    if (CommandStatusInit()) {
                        return(TRUE);
                    }
                    CommandStoreCleanup();            
                }    
                CommandFetchCleanup();
            }
            FreeReturnValueIndex(Imap.command.returnValueIndex);
        }
        HulaKeywordIndexFree(Imap.command.index);    
    }
    return(FALSE);
    
}

static BOOL
FreeImapGlobals()
{
    CommandStatusCleanup();
    CommandStoreCleanup();
    CommandFetchCleanup();
    FreeReturnValueIndex(Imap.command.returnValueIndex);
    HulaKeywordIndexFree(Imap.command.index);
    return(TRUE);
}



static void 
IMAPServer(void *unused)
{
    int ccode;
    unsigned long j;
    XplThreadID oldTGID;
    Connection *conn;
    ImapClient *client;
    XplThreadID id = 0;

    XplRenameThread(XplGetThreadID(), "IMAP Server");
    XplSafeIncrement(Imap.server.active);

    while (!Imap.exiting) {
        if (ConnAccept(Imap.server.conn, &conn) != -1) {
            if (!Imap.exiting) {
                if (!Imap.server.disabled) {
                    if (XplSafeRead(Imap.client.threads.inUse) < (long)Imap.client.threads.max) {
                        client = GetIMAPConnection();
                        if (client) {
                            client->conn = conn;
                    
                            XplBeginCountedThread(&id, HandleConnection, IMAP_STACK_SIZE, client, ccode, Imap.client.threads.inUse);
                            if (ccode == 0) {
                                continue;
                            }
                    
                            ReturnIMAPConnection(client);
                        }
                
                        ConnSend(conn, "* BYE IMAP server out of memory\r\n", strlen("* BYE IMAP server out of memory\r\n"));
                        ConnClose(conn, 1);
                        ConnFree(conn);
                        continue;
                    }

                    ConnSend(conn, "* BYE IMAP connection limit for this server reached\r\n", strlen("* BYE IMAP connection limit for this server reached\r\n"));
                    ConnClose(conn, 1);
                    ConnFree(conn);
                    continue;
                }

                ConnSend(conn, "* BYE IMAP access to this server currently disabled\r\n", strlen("* BYE IMAP access to this server currently disabled\r\n"));
                ConnClose(conn, 1);
                ConnFree(conn);
                continue;
            }
            
            ConnSend(conn, "* BYE IMAP server shutting down\r\n", strlen("* BYE IMAP server shutting down\r\n"));
            ConnClose(conn, 1);
            ConnFree(conn);
            break;
        }

        switch (errno) {
            case ECONNABORTED:
#ifdef EPROTO
            case EPROTO: 
#endif
            case EINTR: {
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, NULL, NULL, errno, 0, NULL, 0);
                continue;
            }

            case EIO: {
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, NULL, NULL, errno, 2, NULL, 0);
                XplConsolePrintf("IMAPD: Server exiting after an accept() failure with an errno: %d\n", errno);

                break;
            }
        }

        break;
    }

/*    Shutting down    */
    Imap.exiting = TRUE;

    oldTGID = XplSetThreadGroupID(Imap.server.threadGroupId);

    if (Imap.server.conn->socket != -1) {
        IPclose(Imap.server.conn->socket);
        Imap.server.conn->socket = -1;
    }

    if(Imap.server.ssl.conn->socket != -1) {
        IPclose(Imap.server.ssl.conn->socket);
        Imap.server.ssl.conn->socket = -1;
    }

    /*    Management Client Shutdown    */
    ManagementShutdown();
    for (j = 0; (ManagementState() != MANAGEMENT_STOPPED) && (j < 60); j++) {
        XplDelay(1000);
    }

    /*    Wake up the children and set them free!    */
    /* fixme - SocketShutdown; */

    /*    Wait for the our siblings to leave quietly.    */
    for (j = 0; (XplSafeRead(Imap.server.active) > 1) && (j < 60); j++) {
        XplDelay(1000);
    }

    if (XplSafeRead(Imap.server.active) > 1) {
        XplConsolePrintf("\rIMAPD: %d server threads outstanding; attempting forceful unload.\r\n", XplSafeRead(Imap.server.active) - 1);
    }

    XplConsolePrintf("\rIMAPD: Shutting down %d client threads\r\n", XplSafeRead(Imap.client.threads.inUse));

    /*    Make sure the kids have flown the coop.    */
    for (j = 0; XplSafeRead(Imap.client.threads.inUse) && (j < 3 * 60); j++) {
        XplDelay(1000);
    }

    if (XplSafeRead(Imap.client.threads.inUse)) {
        XplConsolePrintf("IMAPD: %d threads outstanding; attempting forceful unload.\r\n", XplSafeRead(Imap.client.threads.inUse));
    }

    /* Cleanup SSL */
    XplConsolePrintf("IMAPD: Removing SSL data\r\n");
    if (Imap.server.ssl.context) {
        ConnSSLContextFree(Imap.server.ssl.context);
    }
    ERR_free_strings();
    ERR_remove_state(0);
    EVP_cleanup();
    XPLCryptoLockDestroy();

    LoggerClose(Imap.logHandle);
    Imap.logHandle = NULL;

    MsgShutdown();
//    MDBShutdown();

    CONN_TRACE_SHUTDOWN();
    ConnShutdown();
        
    FreeImapGlobals();
    MemPrivatePoolFree(IMAPConnectionPool);
    MemoryManagerClose(MSGSRV_AGENT_IMAP);

    XplConsolePrintf("IMAPD: Shutdown complete\r\n");

    XplSignalLocalSemaphore(Imap.server.mainSemaphore);
    XplWaitOnLocalSemaphore(Imap.server.shutdownSemaphore);

    XplCloseLocalSemaphore(Imap.server.shutdownSemaphore);
    XplCloseLocalSemaphore(Imap.server.mainSemaphore);

    XplSetThreadGroupID(oldTGID);

    return;
}

static void 
IMAPSSLServer(void *unused)
{
    ImapClient *client;
    unsigned char *message;
    unsigned long messageLen;
    XplThreadID id = 0;
    int ccode;
    Connection *conn;
    
    XplRenameThread(XplGetThreadID(), "IMAP SSL Server");

    while (!Imap.exiting) {
        if (ConnAccept(Imap.server.ssl.conn, &conn) != -1) {
            conn->ssl.enable = TRUE;
            if (!Imap.exiting) {
                if (!Imap.server.disabled) {
                    if (XplSafeRead(Imap.client.threads.inUse) < (long)Imap.client.threads.max) {
                        client = GetIMAPConnection();
                        if (client) {
                            client->conn = conn;
// Imap.client.threads.inUse
                            XplBeginCountedThread(&id, HandleConnection, IMAP_STACK_SIZE, client, ccode, (Imap.client.threads.inUse));
                            if (ccode == 0) {
                                continue;
                            }

                            ReturnIMAPConnection(client);
                            message = "* BYE IMAP server unable to create threads\r\n";
                            messageLen = strlen("* BYE IMAP server unable to create threads\r\n");
                        } else {
                            message = "* BYE IMAP server out of memory\r\n";
                            messageLen = strlen("* BYE IMAP server out of memory\r\n");
                        }
                    } else {
                        message = "* BYE IMAP connection limit for this server reached\r\n";
                        messageLen = strlen("* BYE IMAP connection limit for this server reached\r\n");
                    }
                } else {
                    message = "* BYE IMAP access to this server currently disabled\r\n";
                    messageLen = strlen("* BYE IMAP access to this server currently disabled\r\n");
                }

                if (ConnNegotiate(conn, Imap.server.ssl.context)) {
                    if (ConnWrite(conn, message, messageLen) != -1) {
                        ConnFlush(conn);
                    }
                    ConnClose(conn, 1);
                    ConnFree(conn);
                    continue;
                } 

                ConnSend(conn, message, messageLen);
                ConnClose(conn, 1);
                ConnFree(conn);
                continue;
            }

            break;
        }

        switch (errno) {
            case ECONNABORTED:
#ifdef EPROTO
            case EPROTO:
#endif
            case EINTR: {
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, "SSL Server", NULL, errno, 1, NULL, 0);
                continue;
            }

            case EIO: {
                Imap.exiting = TRUE;
                XplConsolePrintf("IMAPD: SSL Server exiting after an accept() failure with an errno: %d\n", errno);
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ERROR, 0, "SSL Server", NULL, errno, 3, NULL, 0);

                break;
            }

            default: {
                XplConsolePrintf("IMAPD: SSL Server exiting after an accept() failure with an errno: %d\n", errno);
                LoggerEvent(Imap.logHandle, LOGGER_SUBSYSTEM_GENERAL, LOGGER_EVENT_ACCEPT_FAILURE, LOG_ALERT, 0, "SSL Server", NULL, errno, 1, NULL, 0);

                break;
            }
        }

        break;
    }

    XplConsolePrintf("IMAPD: Server SSL listener done.\r\n");
    
    XplSafeDecrement(Imap.server.active);
    
    if (!Imap.exiting) {
        raise(SIGTERM);
    }
    
    return;
}


static BOOL
ReadConfiguration(void)
{
    MDBValueStruct *Config;
    unsigned long i;
    unsigned char *ptr;
    time_t   gm_time_of_day, time_of_day;
    struct tm  gm, ltm;

    tzset();
    time(&time_of_day);
    time(&gm_time_of_day);
    gmtime_r(&gm_time_of_day, &gm);
    localtime_r(&time_of_day, &ltm);
    gm_time_of_day=mktime(&gm);

    Config = MDBCreateValueStruct(Imap.directory.handle, MsgGetServerDN(NULL));

    if (MDBRead(MSGSRV_AGENT_IMAP, MSGSRV_A_CONFIGURATION, Config)>0) {
        for (i=0; i<Config->Used; i++) {
            if (XplStrNCaseCmp(Config->Value[i], "MaxLoad:", 8)==0) {
                Imap.client.threads.max=atol(Config->Value[i]+8);
            } else if (XplStrNCaseCmp(Config->Value[i], "ACLEnabled:", 11) == 0) {
                if (atol(Config->Value[i] + 11)) {
                    Imap.command.capability.acl.enabled = TRUE;
                } else {
                    Imap.command.capability.acl.enabled = FALSE;
                    Imap.command.capability.len = sprintf(Imap.command.capability.message, "%s\r\n", "* CAPABILITY IMAP4 IMAP4rev1 AUTH=LOGIN NAMESPACE XSENDER");
                    Imap.command.capability.ssl.len = sprintf(Imap.command.capability.ssl.message, "%s\r\n", "* CAPABILITY IMAP4 IMAP4rev1 AUTH=LOGIN NAMESPACE STARTTLS XSENDER LOGINDISABLED");
                }
            } else if (XplStrNCaseCmp(Config->Value[i], "Port=", 5) == 0) {
                Imap.server.port = (unsigned short)atol(Config->Value[i] +5);
            } else if (XplStrNCaseCmp(Config->Value[i], "SecurePort=", 5) == 0) {
                Imap.server.ssl.port = (unsigned short)atol(Config->Value[i] +11);
            }
        }
    }
    MDBFreeValues(Config);

    if (MDBRead(MSGSRV_SELECTED_CONTEXT, MSGSRV_A_SSL_OPTIONS, Config)) {
        Imap.server.ssl.config.options=atoi(Config->Value[0]);
        Imap.server.ssl.config.options &= ~SSL_USE_CLIENT_CERT;
        if (Imap.server.ssl.config.options & (SSL_ALLOW_SSL2 | SSL_ALLOW_SSL3)) {
	        Imap.server.ssl.enable = TRUE;
        }
    }
    MDBFreeValues(Config);

    if (MDBRead(MSGSRV_SELECTED_CONTEXT, MSGSRV_A_OFFICIAL_NAME, Config)) {
        if ((Config->Value[0]) && (strlen(Config->Value[0]) < sizeof(Imap.server.hostname))) {
      
            strcpy(Imap.server.hostname, Config->Value[0]);
        } else {
            gethostname(Imap.server.hostname, sizeof(Imap.server.hostname));
        }
  
    } else {
        gethostname(Imap.server.hostname, sizeof(Imap.server.hostname));
    }
    Imap.server.hostnameLen = strlen(Imap.server.hostname);
    MDBFreeValues(Config);

    if (MDBRead(MSGSRV_SELECTED_CONTEXT, MSGSRV_A_POSTMASTER, Config)) {
        ptr=strchr(Config->Value[0], '.');
        if (ptr) {
            *ptr='\0';
        }
        if (strlen(Config->Value[0]) < sizeof(Imap.server.postmaster)) {
            strcpy(Imap.server.postmaster, Config->Value[0]);
        }
    }
    MDBFreeValues(Config);

    MDBDestroyValueStruct(Config);

    return(TRUE);
}

XplServiceCode(IMAPShutdownSigHandler)

int
XplServiceMain(int argc, char *argv[])
{
    int   ccode;
    XplThreadID id=0;

    if (XplSetEffectiveUser(MsgGetUnprivilegedUser()) < 0) {
        XplConsolePrintf("hulaimap: Could not drop to unprivileged user '%s', exiting.\n", MsgGetUnprivilegedUser());
        return 1;
    }

    XplSignalHandler(IMAPShutdownSigHandler);

    if (MemoryManagerOpen(MSGSRV_AGENT_IMAP) == TRUE) {
        IMAPConnectionPool = MemPrivatePoolAlloc("IMAP Connections", sizeof(ImapClient), 0, 3072, TRUE, FALSE, IMAPConnectionAllocCB, IMAPConnectionFreeCB, NULL);
        if (IMAPConnectionPool != NULL) {
            XplOpenLocalSemaphore(Imap.server.mainSemaphore, 0);
            XplOpenLocalSemaphore(Imap.server.shutdownSemaphore, 1);
        } else {
            MemoryManagerClose(MSGSRV_AGENT_IMAP);

            XplConsolePrintf("IMAPD: Unable to create command buffer pool; shutting down.\r\n");
            return(-1);
        }
    } else {
        XplConsolePrintf("IMAPD: Unable to initialize memory manager; shutting down.\r\n");
        return(-1);
    }

    InitializeImapGlobals();

    ConnStartup(CONNECTION_TIMEOUT, TRUE);

    MDBInit();
    Imap.directory.handle = (MDBHandle)MsgInit();
    if (Imap.directory.handle == NULL) {
        XplBell();
        XplConsolePrintf("\rIMAPD: Invalid directory credentials; exiting!\n");
        XplBell();

        MemoryManagerClose(MSGSRV_AGENT_IMAP);

        return(-1);
    }

    NMAPInitialize(Imap.directory.handle);
    CMInitialize(Imap.directory.handle, "IMAP");

    Imap.logHandle = LoggerOpen("hulaimap");
    if (Imap.logHandle != NULL) {
        ;
    } else {
        XplConsolePrintf("IMAPD: Unable to initialize Nsure Audit.  Logging disabled.\r\n");
    }

    ReadConfiguration();
    CONN_TRACE_INIT((char *)MsgGetWorkDir(NULL), "imap");
    // CONN_TRACE_SET_FLAGS(CONN_TRACE_ALL); /* uncomment this line and pass '--enable-conntrace' to autogen to get the agent to trace all connections */

    if (ServerSocketInit() < 0) {
        XplConsolePrintf("hulaimap: Imap.exiting\n");
        return -1;
    }

    if (Imap.server.ssl.enable) {
        if (!ServerSocketSSLInit()) {
        
            Imap.server.ssl.config.method = SSLv23_server_method;
            Imap.server.ssl.config.mode = SSL_MODE_AUTO_RETRY;
            Imap.server.ssl.config.cipherList = NULL;
            Imap.server.ssl.config.certificate.type = SSL_FILETYPE_PEM;
            Imap.server.ssl.config.certificate.file = MsgGetTLSCertPath(NULL);
            Imap.server.ssl.config.key.type = SSL_FILETYPE_PEM;
            Imap.server.ssl.config.key.file = MsgGetTLSKeyPath(NULL);
        
            Imap.server.ssl.context = ConnSSLContextAlloc(&(Imap.server.ssl.config));
            if (Imap.server.ssl.context) {
                XplBeginCountedThread(&id, IMAPSSLServer, IMAP_STACK_SIZE, NULL, ccode, Imap.server.active);
            } else {
                Imap.server.ssl.enable = FALSE;
            }
        } else {
            Imap.server.ssl.enable = FALSE;
        }
    }

    /* Done binding to ports, drop privs permanently */
    if (XplSetRealUser(MsgGetUnprivilegedUser()) < 0) {
        XplConsolePrintf("hulaimap: Could not drop to unprivileged user '%s', exiting\n", MsgGetUnprivilegedUser());
        return 1;
    }

    /* Management Client Startup */
    if ((ManagementInit(MSGSRV_AGENT_IMAP, Imap.directory.handle)) 
        && (ManagementSetVariables(IMAPManagementVariables, sizeof(IMAPManagementVariables) / sizeof(ManagementVariables))) 
        && (ManagementSetCommands(IMAPManagementCommands, sizeof(IMAPManagementCommands) / sizeof(ManagementCommands)))) {
        XplBeginThread(&id, ManagementServer, DMC_MANAGEMENT_STACKSIZE, NULL, ccode);
    }


    XplStartMainThread(PRODUCT_SHORT_NAME, &id, IMAPServer, 8192, NULL, ccode);

    XplUnloadApp(XplGetThreadID());
    return(0);
}
