using System;
using System.Xml;
using System.IO;
using System.Collections;
using Hula.ICal;

namespace Hula.Cal {

//<summary>
// Provides ICalendar subsetting as defined in the CalDav spec
//</summary>
public class CalFilterXml : CalFilter {
    bool includeHeader = true;
    public bool IncludeHeader {
	get { return includeHeader; } 
	set { includeHeader = value; }
    }

    public CalFilterXml(XmlNode node) : base(node) {
    }
    
    public CalFilterXml() : base() {
    }

    protected override void StartDocument(TextWriter writer) {
	if (includeHeader) {
	    writer.WriteLine("<?xml version=\"1.0\" encoding=\"utf-8\" ?>");
	}
	writer.WriteLine("<ical:iCalendar xmlns:ical='http://www.ietf.org/internet-drafts/draft-ietf-calsch-many-xcal-01.txt'>");
    }

    protected override void EndDocument(TextWriter writer) {
	writer.WriteLine("</ical:iCalendar>");
    }

    protected override void StartComponent(TextWriter writer, ICalComponent component) {
	writer.Write("<ical:");
	writer.Write(component.ComponentKind.ToLower());
	writer.WriteLine(">");
    }

    protected override void WriteProperty(TextWriter writer, ICalComponent component, ICalProperty property) {
	if (component.ComponentKind != "") {
	    string xmlName = property.Name.ToLower();
	    writer.Write("<ical:");
	    writer.Write(xmlName);
	    // FIXME: do attributes

	    ICalParameter param = property.GetFirstParameter();
	    while (param != null) {
		writer.Write(" ");
		writer.Write(param.Name.ToLower());
		writer.Write("=\"");
		writer.Write(param.Value);
		writer.Write('"');
		param = property.GetNextParameter();
	    }

	    writer.Write(">");
	    writer.Write(property.Value);
	    writer.Write("</ical:");
	    writer.Write(xmlName);
	    writer.WriteLine(">");
	}
    }

    protected override void EndComponent(TextWriter writer, ICalComponent component) {
	writer.Write("</ical:");
	writer.Write(component.ComponentKind.ToLower());
	writer.WriteLine(">");
    }
}
}
