/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Runtime.InteropServices;
using System.Text;

namespace Hula.CalCmd 
{


public class InvalidCommandException : System.Exception 
{
}


public enum CommandType {
    Error = 0,
    Query,
    QueryConflicts,
    NewAppt,
    Help,
};


[StructLayout(LayoutKind.Sequential, Size=300)]
struct Command
{
    /* static members */
    private static long Epoch = (new DateTime(1970, 1, 1, 0, 0, 0).Ticks);
    /* hack to force memmgr lib to load, see comment in CalCmdRaw */
    private static int Hackity = CalCmdRaw.MemoryManagerOpen("calcmd-foo");

    
    /* properties */
    public CommandType Type {
        get { return (CommandType) CalCmdRaw.GetType(this); }
        set { CalCmdRaw.SetType(ref this, (int) value); }
    }

    public DateTime Begin {
        get { return new DateTime(10000000 * CalCmdRaw.GetBegin(this) + Epoch); }
    }

    public DateTime End {
        get { return new DateTime(10000000 * CalCmdRaw.GetEnd(this) + Epoch); }
    }

    public string Data {
        get { return (string) CalCmdRaw.GetData(this); }
        set { CalCmdRaw.SetData(ref this, value); }
    }


    /* constructor */
    public
    Command(string text)
    {
        CalCmdRaw.SetDebugMode (1);
        if (0 == CalCmdRaw.ParseCommand(text, out this)) {
            throw new ArgumentException ("Couldn't parse command.");
        }
    }


    /* methods */
    public int 
    ProcessNewAppt(out string result) 
    {
        StringBuilder buf;
        int retval;

        buf = new StringBuilder(1024);
        retval = CalCmdRaw.ProcessNewAppt(ref this, buf, buf.Capacity, 
                                          Guid.NewGuid().ToString());
        result = buf.ToString();
        return retval;
    }
    
    
    public void
    PrintConciseDateTimeRange(StringBuilder buffer)
    {
        Command.PrintConciseDateTimeRange(this.Begin, this.End, buffer);
    }
    

    /* static methods */
    public static void 
    PrintConciseDateTimeRange(DateTime begin, DateTime end, StringBuilder buffer)
    {
        StringBuilder buf = new StringBuilder(64);

        CalCmdRaw.PrintConciseDateTimeRange((begin.Ticks - Epoch) / 10000, 
                                            (end.Ticks - Epoch) / 10000,
                                            buf, buf.Capacity);
        buffer.Append(buf);
    }
};



class CalCmdRaw 
{
    const string CalCmdLib = "hulacalcmd";
    const string MemMgrLib = "hulamemmgr";

    [DllImport(CalCmdLib, EntryPoint="CalCmdParseCommand", 
                          ThrowOnUnmappableChar = false)]

    public static extern int 
    ParseCommand (string cmdstr, 
                  out Command cmd);


    [DllImport(CalCmdLib, EntryPoint="CalCmdProcessNewAppt")]

    public static extern int 
    ProcessNewAppt (Command cmd, 
                    out string icalbuf, 
                    int size, 
                    string uid);
        
    [DllImport(CalCmdLib, EntryPoint="CalCmdProcessQuery")]


    public static extern int
    ProcessQuery (Command cmd, 
                  out string buf, 
                  int size);
    

    [DllImport(CalCmdLib, EntryPoint="CalCmdProcessNewAppt")]

    public static extern int
    ProcessNewAppt (ref Command cmd, 
                    StringBuilder icalbuf, 
                    int size, 
                    string uid);


    [DllImport(CalCmdLib, EntryPoint="CalCmdGetType")]
    public static extern int
    GetType(Command cmd);

    
    [DllImport(CalCmdLib, EntryPoint="CalCmdSetType")]

    public static extern void
    SetType(ref Command cmd, int value);


    [DllImport(CalCmdLib, EntryPoint="CalCmdGetData")]
    private static extern IntPtr
    GetDataHelper(Command cmd);

    public static string
    GetData(Command cmd)
    {
        return Marshal.PtrToStringAnsi(GetDataHelper(cmd));
    }


    [DllImport(CalCmdLib, EntryPoint="CalCmdSetData", ThrowOnUnmappableChar=false)]

    public static extern void
    SetData(ref Command cmd, string data);



    [DllImport(CalCmdLib, EntryPoint="CalCmdGetBegin")]

    public static extern long
    GetBegin([MarshalAs(UnmanagedType.LPStruct)]Command cmd);


    [DllImport(CalCmdLib, EntryPoint="CalCmdGetEnd")]

    public static extern long
    GetEnd(Command cmd);


    [DllImport(CalCmdLib, EntryPoint="CalCmdSetDebugMode")]
    public static extern int
    SetDebugMode(int flag);


    [DllImport(CalCmdLib, EntryPoint="CalCmdPrintConciseDateTimeRange")]

    public static extern int
    PrintConciseDateTimeRange(long beginMillisLocal, long endMillisLocal, 
                              StringBuilder buffer, int size);



    /* following the lead of Mdb.cs, it seems I need to import these functions 
       for libcalcmd to work properly (on Linux, at least) */

    [DllImport(MemMgrLib)]
    public static extern int MemoryManagerOpen (string agentName);

    [DllImport(MemMgrLib)]
    public static extern int MemoryManagerClose (string agentName);
    
}

}

