/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/

using System;
using System.Collections;
using System.IO;

using Hula.Store;

namespace Hula.Importing {

public class VCardImporter : Importer
{
    TextReader Reader;

    public VCardImporter (string filename, string contentType, Stream stream)
    {
	Reader = new StreamReader (stream);
    }

    public override IList ImportTo (IStore store, string folder)
    {
	ArrayList guids = new ArrayList ();
	string s;
	string guid, etag;
	Hashtable props = new Hashtable ();

	do {
	    for (s = Reader.ReadLine ();
		 s != null && String.Compare (s, "begin:vcard", true) != 0;
		 s = Reader.ReadLine ()) {
		// Console.Error.WriteLine (" >>> no begin: {0}", s);
		/* noop */ ;
	    }

	    if (s == null) {
		break;
	    }

	    using (StringWriter writer = new StringWriter ()) {
		writer.WriteLine (s);
		for (s = Reader.ReadLine ();
		     s != null && String.Compare (s, "end:vcard", true) != 0;
		     s = Reader.ReadLine ()) {
		    // Console.Error.WriteLine (" >>> adding data: {0}", s);
		    writer.WriteLine (s);
		}

		if (s == null) {
		    break;
		}

		writer.WriteLine (s);

		store.Add (folder, props, new StringReader (writer.ToString ()),
			   "text/x-vcard", out guid, out etag);

		// Console.Error.WriteLine (" >>> added: {0} ({1})", guid, etag);


		guids.Add (guid);
	    }
	} while (true);

	return guids;
    }
}

}
