/****************************************************************************
 *
 * Copyright (c) 2005 Novell, Inc.
 * All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2.1 of the GNU Lesser General Public
 * License as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, contact Novell, Inc.
 *
 * To contact Novell about this file by physical or electronic mail,
 * you may find current contact information at www.novell.com
 *
 ****************************************************************************/
using System;
using System.Runtime.InteropServices;
using System.Text;

using log4net;

namespace Hula.Mdb
{
public class MsgApi
{
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    static MsgApi ()
    {
        MsgApiRaw.Init();
    }

    public static string NmapChallenge (string salt)
    {
        // we're getting back a 32-character md5 hash, null-terminated
        StringBuilder buf = new StringBuilder(33);

        MsgApiRaw.NmapChallenge(salt, buf, buf.Capacity);

        return buf.ToString();
    }

    public static string FindUserNmap (string user)
    {
        StringBuilder buf = new StringBuilder(16);

        MsgApiRaw.MsgFindUserNmap(user, buf, buf.Capacity);

        log.DebugFormat("Finding NMAP store for {0}... {1}", user, buf);

        return buf.ToString();
    }

    public static bool FindObject(string user, out string dn) {
	StringBuilder buf = new StringBuilder(256);
	bool ret = MsgApiRaw.FindObject(user, buf, null, IntPtr.Zero, IntPtr.Zero);
	dn = buf.ToString();

	return ret;
    }
}

class MsgApiRaw
{
    const string MsgApiLib = "hulamsgapi";

    [DllImport (MsgApiLib, EntryPoint="MsgNmapChallenge")]
    public static extern void NmapChallenge (string response,
                                             StringBuilder buf, int len);

    [DllImport (MsgApiLib, EntryPoint="MsgInit")]
    public static extern void Init ();

    [DllImport(MsgApiLib, EntryPoint="MsgFindObject")]
    public static extern bool FindObject(string user, StringBuilder dn, string type, IntPtr nmap, IntPtr valueStruct);

    [DllImport(MsgApiLib, EntryPoint="MsgFindUserNmap")]
    public static extern bool MsgFindUserNmap(string user, StringBuilder nmap, int nmap_len);

}
}
