using System;
using System.Collections;
using System.Web;
using System.IO;

using Hula.Nmap;
using Hula.Mdb;
using Hula.Util;

using log4net;

namespace Hula.Queue {

public class QueueService : IHttpHandler  {
    private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

    static QueueService() {
	Logging.Configure();
	log.Info("started logging");
    }

    public bool IsReusable {
	get { return true; }
    }
    private void HandleOptions(HttpContext context, string user) {
	context.Response.StatusCode = 200;
	context.Response.AddHeader("Allow", "OPTIONS, PUT");
    }

    private void AddRecipients(NmapClient nmap, bool local, string header) {
	if (header == null) {
	    return;
	}

	string[] recipients = null;
	recipients = header.Split(new char[] {',', ' '});
	foreach (string recipient in recipients) {
	    string[] pieces = recipient.Split(new char[] {'|'});
	    if (pieces.Length > 1) {
		if (local) {
		    nmap.Auth.QSTOR_LOCAL(pieces[0], pieces[1], 0);
		} else {
		    nmap.Auth.QSTOR_TO(pieces[0], pieces[1], 0);
		}
	    } else {
		if (local) {
		    nmap.Auth.QSTOR_LOCAL(recipient, recipient, 0);
		} else {
		    nmap.Auth.QSTOR_TO(recipient, recipient, 0);
		}
	    }
	}
    }

    private void HandlePut(HttpContext context, string user) {
	log.DebugFormat ("Sending message for {0}", user);

	using (NmapClient nmap = new NmapClient(user)) {
	    nmap.Auth.QCREA();
	    try {
		nmap.Auth.QSTOR_FROM(user, user);
		
		AddRecipients(nmap, true, 
			      context.Request.Headers["X-Hula-LocalRecipients"]);
		AddRecipients(nmap, false, 
			      context.Request.Headers["X-Hula-Recipients"]);
		
		nmap.Auth.QSTOR_MESSAGE(new StreamReader (context.Request.InputStream));

		nmap.Auth.QRUN();
	    } catch (Exception e) {
		nmap.Auth.QABRT();
		throw e;
	    }
	}
    }

    public void ProcessRequest (HttpContext context) {
	string user = context.User.Identity.Name;


        string method = context.Request.Headers["X-Hula-HttpMethod"];
        if (method == null) {
            method = context.Request.HttpMethod;
        }
	log.DebugFormat("{0} {1}", 
			method,
			context.Request.Path);

	try {
	    switch (method) {
	    case "OPTIONS" :
		HandleOptions(context, user);
		break;
	    case "PUT" :
		HandlePut(context, user);
		break;
	    default :
		throw new HttpException (501, "Not Implemented: " + method);
	    }
	} catch (NmapException e) {
	    Exception newE = Hula.Nmap.NmapClient.NmapToHttpException(e);
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   400);
	    log.Warn(newE);
	    throw newE;
	} catch (HttpException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   e.GetHttpCode());
	    log.Warn(e);
	    throw e;
	} catch (NotImplementedException e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   501);
	    log.Warn(e);
	    throw new HttpException(501, "Not Implemented", e);
	} catch (Exception e) {
	    log.InfoFormat("{0} {1}: {2}", 
			   method,
			   context.Request.Path,
			   500);
	    log.Warn(e);
	    throw e;
	}

	log.InfoFormat("{0} {1}: {2}", 
		       method,
		       context.Request.Path,
		       context.Response.StatusCode);
    }
}
}
