using System;
using System.Collections;
using System.Web;
using System.IO;
using System.Xml;
using Lucene.Net;
using Lucene.Net.Analysis;
using Lucene.Net.Analysis.Standard;
using Lucene.Net.Documents;
using Lucene.Net.Index;
using Lucene.Net.Store;
using Lucene.Net.Search;
using Lucene.Net.QueryParsers;

using Hula.Util;
using Hula.Indexing;

namespace Hula.Search {

//<summary> 
//This is an internal-only search service
//</summary>
public class SearchService : IHttpHandler  {

    public static void DecomposeUrl(string url,
				    out string owner,
				    out string store,
				    out string path)
    {
	owner = store = path = null;

	url = url.Substring(1);

	int i;
	i = url.IndexOf('/');

	if (i > -1) {
	    owner = url.Substring(0, i);
	    url = url.Substring(i + 1);
	} else {
	    owner = url;
	    store = null;
	    path = null;
	    return;
	}

	i = url.IndexOf('/');

	if (i > 0) {
	    store = url.Substring(0, i);
	    path = url.Substring(i);
	} else {
	    store = url;
	    path = null;
	}
    }    

    public bool IsReusable {
	get { return true; }
    }

    private string GetLocalPath(HttpRequest request, string path) {
	if (path.StartsWith (request.ApplicationPath)) {
	    return path.Substring(request.ApplicationPath.Length).TrimEnd(new char[] { '/' });
	} else {
	    return path.TrimEnd(new char[] { '/' });
	}
    }

    private void HandleOptions(HttpContext context, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);

	context.Response.StatusCode = 200;
	context.Response.AddHeader("Allow", "GET");
    }

    private void FormatResultsGuid(HttpContext context, Hits hits) {
	context.Response.ContentType = "text/plain";
	
	int count = hits.Length();
	
	for (int i = 0; i < count; i++) {
	    Document doc = hits.Doc(i);
	    context.Response.Output.WriteLine(doc.GetField("guid").StringValue());
	}
    }

    private void FormatResultsCal(HttpContext context, Hits hits) {
	context.Response.ContentType = "text/plain";
	
	int count = hits.Length();
	
	for (int i = 0; i < count; i++) {
	    Document doc = hits.Doc(i);
	    context.Response.Output.Write(doc.GetField("guid").StringValue());
	    context.Response.Output.Write('|');	    
	    context.Response.Output.Write(doc.GetField("hula.start-time").StringValue());
	    context.Response.Output.Write('|');	    
	    context.Response.Output.Write(doc.GetField("hula.end-time").StringValue());
	    context.Response.Output.Write('|');
	    context.Response.Output.Write(doc.GetField("ical.summary").StringValue());
	}
    }

    private void HandleGet(HttpContext context, string href) {
	string owner, storeName, path;

	DecomposeUrl(href, out owner, out storeName, out path);

	string query = context.Request.QueryString["query"];
	string format = context.Request.QueryString["format"];
	string sortField = context.Request.QueryString["sortkey"];
	string sortOrder = context.Request.QueryString["sortorder"];

	using (LuceneIndex idx = IndexFactory.GetIndex(owner)) {
	    QueryParser parser = new QueryParser("body", new HulaAnalyzer());
	    
	    Query q = parser.Parse(query);
	    Hits hits;
	    if (sortField != null) {
		bool reversed = false;
		if (sortOrder != null && sortOrder[0] == 'd') {
		    reversed = true;
		}
		Sort sort = new Sort(new SortField(sortField, SortField.STRING, reversed));
		hits = idx.Searcher.Search(q, sort);
	    } else {
		hits = idx.Searcher.Search(q);
	    }

	    context.Response.StatusCode = 200;
	    if (format == "cal") {
		FormatResultsCal(context, hits);
	    } else {
		FormatResultsGuid(context, hits);
	    }
        }            
    }

    public void ProcessRequest (HttpContext context) {
        string path = GetLocalPath(context.Request, context.Request.Path);

	try {
	    switch (context.Request.HttpMethod) {
	    case "OPTIONS" :
		HandleOptions(context, path);
		break;
	    case "GET" :
		HandleGet(context, path);
		break;
	    }
	} catch (XmlException e) {
	    throw new HttpException(400, "Could not parse request body", e);
	}
    }
}
}
