
using System;
using System.Configuration;
using System.IO;
using System.Security.Principal;
using System.Text;
using System.Web;
using System.Xml;
using Hula.Mdb;
using log4net;

namespace Hula.Util {
    class HulaAuthenticationModule : IHttpModule {
	private static readonly ILog log = LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

	static char[] separator = {':'};
	
	static bool DummyAuth;
	static HulaAuthenticationModule() {
	    DummyAuth = (Environment.GetEnvironmentVariable ("HULA_DUMMY_AUTH") != null);
	}

	Hula.Mdb.Mdb mdb;

	public HulaAuthenticationModule () { }
	
	public virtual void Init (HttpApplication context) {
	    context.AuthenticateRequest += new EventHandler (this.OnAuthenticateRequest);
	    context.EndRequest += new EventHandler (this.OnEndRequest);
	    
	    if (!DummyAuth) {
		mdb = new Hula.Mdb.Mdb(null, null);
	    }
	}

	public virtual void Dispose () {}

	public virtual void OnAuthenticateRequest (object source, EventArgs eventArgs) {
	    HttpApplication app = (HttpApplication) source;
	    string authdata = Authorization (app);

	    if (!AcceptCredentials (app, authdata)) {
		DenyAccess (app);
		return;
	    }
	}

	// We add the WWW-Authenticate header here, so if an authorization
	// fails elsewhere than in this module, we can still request authentication
	// from the client.
	public void OnEndRequest (object source, EventArgs eventArgs) {
	    HttpApplication app = (HttpApplication) source;
	    if (app.Response.StatusCode != 401) {
		return;
	    }

	    app.Response.AppendHeader ("WWW-Authenticate", "Basic realm=\"Hula\"");
	}
	
	protected void DenyAccess (HttpApplication app) {
	    // FIXME: if the user supplied data, we need to read it.
	    // Is there a better way of figuring out whether the 
	    // user supplied data?  (although ideally xsp would 
	    // just deal with this for us)
	    if (app.Context.Request.HttpMethod == "PUT" ||
		app.Context.Request.HttpMethod == "POST" ||
		app.Context.Request.HttpMethod == "REPORT") {
		while (app.Context.Request.InputStream.ReadByte() != -1) {
		}
	    }

	    app.Response.StatusCode = 401;
	    app.Response.StatusDescription = "Access Denied";
	    // Write to response stream as well, to give user visual
	    // indication of error during development
	    app.Response.Write ("401 Access Denied");
	    app.CompleteRequest ();
	}

	protected bool CheckUserPass(string user, string pass) {
	    if (DummyAuth) {
		return user == pass;
	    } else {	    
		string dn;
		MsgApi.FindObject(user, out dn);
		return mdb.VerifyPassword(dn, pass);
	    }
	}

	protected bool AcceptCredentials (HttpApplication app, string authentication) {
            byte[] userpass;
	    if (authentication == null) {
		return false;
	    }

            try {
                userpass = Convert.FromBase64String (authentication);
            } catch (FormatException e) {
                return false;
            }

	    string[] up = Encoding.UTF8.GetString (userpass).Split (separator, 2);
            if (up.Length != 2) {
                return false;
            }

	    string username = up [0];
	    string password = up [1];
	    
	    if (!CheckUserPass(username, password)) {
		log.WarnFormat("Denying access for user {0}", username);
		return false;
	    }

	    string[] roles = new string[1];
	    roles[0] = "user";
	    app.Context.User = new GenericPrincipal (new GenericIdentity (username, "Basic"), roles);
	    return true;
	}

	protected string Authorization (HttpApplication app) {
            string basic = "BASIC";
	    string autz = app.Request.Headers ["Authorization"];

	    if (autz != null && autz.ToUpper().StartsWith(basic)) {
                return autz.Substring (basic.Length + 1);
	    }

            if (app.Request.ContentType.StartsWith ("multipart/form-data") &&
                app.Request.Form != null) {
                return app.Request.Form["HulaAuthToken"];
            }

	    return null;
	}
    }
}
