/*
 * <Novell-copyright>
 * Copyright (c) 2001 Novell, Inc. All Rights Reserved.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of version 2 of the GNU General Public License
 * as published by the Free Software Foundation.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, contact Novell, Inc.
 * 
 * To contact Novell about this file by physical or electronic mail, you 
 * may find current contact information at www.novell.com.
 * </Novell-copyright>
 */


#ifndef MDBP_H
#define MDBP_H

#include <ldap.h>

/* Little helpers */

/*  We'll just use 1 namingContext in our MDBldap.namingContext array,       */
/*  unless/until we decide that we need to support more.                     */
#define MDBLDAP_MAX_NAMING_CONTEXTS 1

enum MDBLDAP_TYPE {
        MDBLDAP_TYPE_NONE,
        MDBLDAP_TYPE_EDIR,
        MDBLDAP_TYPE_OPENLDAP,
        MDBLDAP_TYPE_AD
};

#define MDB_CONTEXT_IS_VALID(ValueStruct)   (ValueStruct->Flags & MDB_FLAGS_CONTEXT_VALID)

/* Bit definitions */
#define MDB_FLAGS_ALLOCATED_HANDLE          (unsigned long)(1<<0)            /* The Init function has allocated the handle */
#define MDB_FLAGS_CONTEXT_VALID             (unsigned long)(1<<1)            /* The Context handle is valid */
#define MDB_FLAGS_CONTEXT_DUPLICATE         (unsigned long)(1<<2)            /* The context handle is duplicated, do not free it */
#define MDB_FLAGS_BASEDN_CHANGED            (unsigned long)(1<<3)            /* Caching uses this, if the DN of a context is changed, this is set */

#define VALUE_ALLOC_SIZE                    20
#define VALUE_BUFFER_SIZE                   (10 * 1024)

#define ERR_NO_SUCH_ENTRY                   0xFFFFFDA7
#define ERR_NO_SUCH_VALUE                   0xFFFFFDA6
#define ERR_NO_SUCH_ATTRIBUTE               0xFFFFFDA5
#define ERR_NO_SUCH_CLASS                   0xFFFFFDA4
#define ERR_NO_SUCH_PARTITION               0xFFFFFDA3
#define ERR_ENTRY_ALREADY_EXISTS            0xFFFFFDA2
#define ERR_NOT_EFFECTIVE_CLASS             0xFFFFFDA1
#define ERR_ILLEGAL_ATTRIBUTE               0xFFFFFDA0
#define ERR_MISSING_MANDATORY               0xFFFFFD9F
#define ERR_ILLEGAL_DS_NAME                 0xFFFFFD9E
#define ERR_ILLEGAL_CONTAINMENT             0xFFFFFD9D
#define ERR_CANT_HAVE_MULTIPLE_VALUES       0xFFFFFD9C
#define ERR_SYNTAX_VIOLATION                0xFFFFFD9B
#define ERR_DUPLICATE_VALUE                 0xFFFFFD9A
#define ERR_ATTRIBUTE_ALREADY_EXISTS        0xFFFFFD99
#define ERR_MAXIMUM_ENTRIES_EXIST           0xFFFFFD98
#define ERR_DATABASE_FORMAT                 0xFFFFFD97
#define ERR_INCONSISTENT_DATABASE           0xFFFFFD96
#define ERR_INVALID_COMPARISON              0xFFFFFD95
#define ERR_COMPARISON_FAILED               0xFFFFFD94
#define ERR_TRANSACTIONS_DISABLED           0xFFFFFD93
#define ERR_INVALID_TRANSPORT               0xFFFFFD92
#define ERR_SYNTAX_INVALID_IN_NAME          0xFFFFFD91
#define ERR_TRANSPORT_FAILURE               0xFFFFFD8F
#define ERR_ALL_REFERRALS_FAILED            0xFFFFFD8E
#define ERR_CANT_REMOVE_NAMING_VALUE        0xFFFFFD8D
#define ERR_OBJECT_CLASS_VIOLATION          0xFFFFFD8C
#define ERR_ENTRY_IS_NOT_LEAF               0xFFFFFD8B
#define ERR_DIFFERENT_TREE                  0xFFFFFD8A
#define ERR_SYSTEM_FAILURE                  0xFFFFFD88
#define ERR_INVALID_ENTRY_FOR_ROOT          0xFFFFFD87
#define ERR_NO_REFERRALS                    0xFFFFFD86
#define ERR_REMOTE_FAILURE                  0xFFFFFD85
#define ERR_UNREACHABLE_SERVER              0XFFFFFD84
#define ERR_NO_CHARACTER_MAPPING            0XFFFFFD82
#define ERR_INCOMPLETE_AUTHENTICATION       0XFFFFFD81
#define ERR_INVALID_CERTIFICATE             0xFFFFFD80
#define ERR_INVALID_REQUEST                 0xFFFFFD7F
#define ERR_INVALID_ITERATION               0xFFFFFD7E
#define ERR_SCHEMA_IS_NONREMOVABLE          0xFFFFFD7D
#define ERR_SCHEMA_IS_IN_USE                0xFFFFFD7C
#define ERR_CLASS_ALREADY_EXISTS            0xFFFFFD7B
#define ERR_BAD_NAMING_ATTRIBUTES           0xFFFFFD7A
#define ERR_INSUFFICIENT_STACK              0xFFFFFD78
#define ERR_INSUFFICIENT_BUFFER             0xFFFFFD77
#define ERR_AMBIGUOUS_CONTAINMENT           0xFFFFFD76
#define ERR_AMBIGUOUS_NAMING                0xFFFFFD75
#define ERR_DUPLICATE_MANDATORY             0xFFFFFD74
#define ERR_DUPLICATE_OPTIONAL              0xFFFFFD73
#define ERR_RECORD_IN_USE                   0xFFFFFD6C
#define ERR_ENTRY_NOT_CONTAINER             0xFFFFFD64
#define ERR_FAILED_AUTHENTICATION           0xFFFFFD63
#define ERR_INVALID_CONTEXT                 0xFFFFFD62
#define ERR_NO_SUCH_PARENT                  0xFFFFFD61
#define ERR_NO_ACCESS                       0xFFFFFD60
#define ERR_INVALID_NAME_SERVICE            0xFFFFFD5E
#define ERR_INVALID_TASK                    0xFFFFFD5D
#define ERR_INVALID_CONN_HANDLE             0xFFFFFD5C
#define ERR_INVALID_IDENTITY                0xFFFFFD5B
#define ERR_DUPLICATE_ACL                   0xFFFFFD5A
#define ERR_ALIAS_OF_AN_ALIAS               0xFFFFFD57
#define ERR_INVALID_RDN                     0xFFFFFD4E
#define ERR_INCORRECT_BASE_CLASS            0xFFFFFD4C
#define ERR_MISSING_REFERENCE               0xFFFFFD4B
#define ERR_LOST_ENTRY                      0xFFFFFD4A
#define ERR_FATAL                           0xFFFFFD45      /*  -699   */
#define ERR_NOT_IMPLEMENTED                 0xFFFFFD36      /*  -714   */
#define ERR_INSUFFICIENT_MEMORY             0xFFFFFF6A      /*  -150   */

typedef struct {
    /* Semi-public; known to the main MDB library */
    unsigned char Description[128];
    struct _MDBInterfaceStruct *Interface;

    /* Private */
	LDAP *bindhandle;
    XplSemaphore sem;
} MDBContextStruct;

typedef MDBContextStruct *MDBHandle;

typedef struct {
    /* Public part, exposed to allow simpler code */
    unsigned char **Value;
    unsigned long Used;
    unsigned long ErrNo;
    struct _MDBInterfaceStruct *Interface;

    /* Private part, not exposed in public header */
    unsigned long Flags;
    unsigned long Allocated;
    unsigned char *BaseDN;
	MDBContextStruct *context;
} MDBValueStruct;

typedef struct {
    BOOL Initialized;

    unsigned int Entry;

    MDBValueStruct *V;
} MDBEnumStruct;

/* this is so we don't try to define MDBValueStruct more than once */
#ifndef MDB_INTERNAL
#define MDB_INTERNAL
#endif

//#include <schema.h>
#include <mdb.h>

#endif /* MDBP_H */
